"""K-means clustering.

Lead author: Hadi Zaatiti.
"""

import logging
from random import randint

from scipy.stats import rv_discrete
from sklearn.base import BaseEstimator, ClusterMixin

import geomstats.backend as gs
from geomstats.learning._template import TransformerMixin
from geomstats.learning.frechet_mean import FrechetMean


class RiemannianKMeans(TransformerMixin, ClusterMixin, BaseEstimator):
    """Class for k-means clustering on manifolds.

    K-means algorithm using Riemannian manifolds.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    n_clusters : int
        Number of clusters (k value of the k-means).
        Optional, default: 8.
    init : str or callable or array-like, shape=[n_clusters, n_features]
        How to initialize cluster centers at the beginning of the algorithm. The
        choice 'random' will select training points as initial cluster centers
        uniformly at random. The choice 'kmeans++' selects cluster centers
        heuristically to improve the convergence rate. When providing an array
        of shape ``(n_clusters, n_features)``, the cluster centers are chosen as the
        rows of that array. When providing a callable, it receives as arguments
        the argument ``X`` to :meth:`fit` and the number of cluster centers
        ``n_clusters`` and is expected to return an array as above.
        Optional, default: 'random'.
    tol : float
        Convergence factor. Convergence is achieved when the difference of mean
        distance between two steps is lower than tol.
        Optional, default: 1e-2.
    max_iter : int
        Maximum number of iterations.
        Optional, default: 100
    verbose : int
        If verbose > 0, information will be printed during learning.
        Optional, default: 0.

    Notes
    -----
    * Required metric methods: `dist`.

    Example
    -------
    Available example on the Poincaré Ball and Hypersphere manifolds
    :mod:`examples.plot_kmeans_manifolds`
    """

    def __init__(
        self,
        space,
        n_clusters=8,
        init="random",
        tol=1e-2,
        max_iter=100,
        verbose=0,
    ):
        self.space = space

        self.n_clusters = n_clusters
        self.init = init
        self.tol = tol
        self.verbose = verbose
        self.max_iter = max_iter

        self.init_cluster_centers_ = None

        self.mean_estimator = FrechetMean(space=space)
        if isinstance(self.mean_estimator, FrechetMean):
            self.mean_estimator.set(max_iter=100, init_step_size=1.0)

        self.cluster_centers_ = None
        self.labels_ = None
        self.inertia_ = None

    def _pick_init_cluster_centers(self, X):
        n_samples = X.shape[0]

        if isinstance(self.init, str):
            if self.init == "kmeans++":
                cluster_centers = [X[randint(0, n_samples - 1)]]
                for i in range(self.n_clusters - 1):
                    dists = gs.array(
                        [
                            self.space.metric.dist(cluster_centers[j], X)
                            for j in range(i + 1)
                        ]
                    )
                    dists_to_closest_cluster_center = gs.amin(dists, axis=0)
                    indices = gs.arange(n_samples)
                    weights = dists_to_closest_cluster_center / gs.sum(
                        dists_to_closest_cluster_center
                    )
                    index = rv_discrete(
                        values=(gs.to_numpy(indices), gs.to_numpy(weights))
                    ).rvs()
                    cluster_centers.append(X[index])
            elif self.init == "random":
                cluster_centers = [
                    X[randint(0, n_samples - 1)] for i in range(self.n_clusters)
                ]
            else:
                raise ValueError(
                    f"Unknown initial cluster centers method '{self.init}'."
                )

            cluster_centers = gs.stack(cluster_centers, axis=0)
        else:
            if callable(self.init):
                cluster_centers = self.init(X, self.n_clusters)
            else:
                cluster_centers = self.init

            if cluster_centers.shape[0] != self.n_clusters:
                raise ValueError("Need as many initial cluster centers as clusters.")

            if cluster_centers.shape[1] != X.shape[1]:
                raise ValueError(
                    "Dimensions of initial cluster centers and "
                    "training data do not match."
                )

        return cluster_centers

    def fit(self, X):
        """Provide cluster centers and data labels.

        Alternate between computing the mean of each cluster
        and labelling data according to the new positions of the cluster centers.

        Parameters
        ----------
        X : array-like, shape=[n_samples, n_features]
            Training data, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        self : object
            Returns self.
        """
        n_samples = X.shape[0]
        if self.verbose > 0:
            logging.info("Initializing...")

        cluster_centers = self._pick_init_cluster_centers(X)
        self.init_cluster_centers_ = gs.copy(cluster_centers)

        dists = [
            gs.to_ndarray(self.space.metric.dist(cluster_centers[i], X), 2, 1)
            for i in range(self.n_clusters)
        ]
        dists = gs.hstack(dists)
        self.labels_ = gs.argmin(dists, 1)

        for index in range(self.max_iter):
            if self.verbose > 0:
                logging.info(f"Iteration {index}...")

            old_cluster_centers = gs.copy(cluster_centers)
            for i in range(self.n_clusters):
                fold = X[self.labels_ == i]

                if len(fold) > 0:
                    self.mean_estimator.fit(fold)
                    cluster_centers[i] = self.mean_estimator.estimate_
                else:
                    cluster_centers[i] = X[randint(0, n_samples - 1)]

            dists = [
                gs.to_ndarray(self.space.metric.dist(cluster_centers[i], X), 2, 1)
                for i in range(self.n_clusters)
            ]
            dists = gs.hstack(dists)
            self.labels_ = gs.argmin(dists, 1)
            dists_to_closest_cluster_center = gs.amin(dists, 1)
            self.inertia_ = gs.sum(dists_to_closest_cluster_center**2)
            cluster_centers_distances = self.space.metric.dist(
                old_cluster_centers, cluster_centers
            )
            if self.verbose > 0:
                logging.info(
                    f"Convergence criterion at the end of iteration {index} "
                    f"is {gs.mean(cluster_centers_distances)}."
                )

            if gs.mean(cluster_centers_distances) < self.tol:
                if self.verbose > 0:
                    logging.info(f"Convergence reached after {index} iterations.")

                break
        else:
            logging.warning(
                f"K-means maximum number of iterations {self.max_iter} reached. "
                "The mean may be inaccurate."
            )

        self.cluster_centers_ = cluster_centers

        return self

    def predict(self, X):
        """Predict the labels for each data point.

        Label each data point with the cluster having the nearest
        cluster center using metric distance.

        Parameters
        ----------
        X : array-like, shape[n_samples, n_features]
            Input data.

        Returns
        -------
        labels : array-like, shape=[n_samples,]
            Array of predicted cluster indices for each sample.
        """
        if self.cluster_centers_ is None:
            raise RuntimeError("fit needs to be called first.")
        dists = gs.stack(
            [
                self.space.metric.dist(cluster_center, X)
                for cluster_center in self.cluster_centers_
            ],
            axis=1,
        )
        dists = gs.squeeze(dists)

        labels = gs.argmin(dists, -1)

        return labels



"""Visualization for Geometric Statistics."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs

AX_SCALE = 1.2


class PoincarePolyDisk:
    """Class used to plot points in the Poincare polydisk."""

    def __init__(self, points=None, coords_type="ball", n_disks=2):
        self.center = gs.array([0.0, 0.0])
        self.points = []
        self.coords_type = coords_type
        self.n_disks = n_disks
        if points is not None:
            self.add_points(points)

    @staticmethod
    def set_ax(ax=None):
        """Define the ax parameters."""
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(ax, xlim=(-ax_s, ax_s), ylim=(-ax_s, ax_s), xlabel="X", ylabel="Y")
        return ax

    def add_points(self, points):
        """Add points to draw."""
        if self.coords_type == "extrinsic":
            points = self.convert_to_poincare_coordinates(points)
        if not isinstance(points, list):
            points = list(points)
        self.points.extend(points)

    def clear_points(self):
        """Clear the points to draw."""
        self.points = []

    @staticmethod
    def convert_to_poincare_coordinates(points):
        """Convert points to poincare coordinates."""
        poincare_coords = points[:, 1:] / (1 + points[:, :1])
        return poincare_coords

    def draw(self, ax, **kwargs):
        """Draw."""
        circle = plt.Circle((0, 0), radius=1.0, color="black", fill=False)
        ax.add_artist(circle)
        points_x = [gs.to_numpy(point[0]) for point in self.points]
        points_y = [gs.to_numpy(point[1]) for point in self.points]
        ax.scatter(points_x, points_y, **kwargs)



"""The ProductPositiveRealsAndComplexPoincareDisks manifold.

The ProductPositiveRealsAndComplexPoincareDisks manifold is defined as a
product manifold of the positive reals manifold and (n-1) complex Poincaré disks.
The positive reals and complex Poincaré disks product has a product metric.
The product metric on the positive reals and complex Poincaré disks product space is
the positive reals metric and (n - 1) complex Poincaré metrics multiplied by constants.
This product manifold can be used to represent Toeplitz HPD matrices.
The ProductPositiveRealsAndComplexPoincareDisks corresponds to
the one-dimensional case of ProductHPDMatricesAndSiegelDisks.
Indeed, PositiveReals is the one-dimensional case of HPDMatrices and
ComplexPoincareDisk is the one-dimensional case of Siegel.
In these one-dimensional manifolds, many simplifications occur compared with
the multidimensional manifolds since matrices commute in dimension 1.


Lead author: Yann Cabanes.

References
----------
.. [Cabanes_2022] Yann Cabanes. Multidimensional complex stationary
    centered Gaussian autoregressive time series machine learning
    in Poincaré and Siegel disks: application for audio and radar
    clutter classification, PhD thesis, tel-03708515, 2022.
    https://theses.hal.science/tel-03708515
.. [Cabanes_CESAR_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon et
    Jérémie Bigot. Unsupervised Machine Learning for Pathological Radar Clutter
    Clustering: the P-Mean-Shift Algorithm, IEEE, C&ESAR 2019, Rennes, France, 2019.
    https://hal.archives-ouvertes.fr/hal-02875430
.. [Cabanes_RADAR_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon et
    Jérémie Bigot. Non-Supervised High Resolution Doppler Machine Learning for
    Pathological Radar Clutter, IEEE, RADAR 2019, Toulon, France, 2019.
    https://hal.archives-ouvertes.fr/hal-02875415
.. [Cabanes_GSI_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon et
    Jérémie Bigot. Toeplitz Hermitian Positive Definite Matrix Machine Learning
    based on Fisher Metric, IEEE, GSI 2019, Toulouse, France, 2019.
    https://hal.archives-ouvertes.fr/hal-02875403
.. [Le_Brigant_2017] Alice Le Brigant. Probability on the spaces of curves and
    the associated metric spaces using information geometry; radar applications,
    PhD thesis, tel-01635258, 2017.
    https://theses.hal.science/tel-01635258
.. [Jeuris_2016] B. Jeuris and R. Vandebril. The Kahler mean of Block-Toeplitz
    matrices with Toeplitz structured blocks, 2016.
    https://epubs.siam.org/doi/pdf/10.1137/15M102112X
.. [Yang_2013] Marc Arnaudon, Frédéric Barbaresco and Le Yang. Riemannian Medians
    and Means With Applications to Radar Signal Processing, IEEE, 2013.
"""

from geomstats.geometry.complex_poincare_disk import ComplexPoincareDisk
from geomstats.geometry.positive_reals import PositiveReals
from geomstats.geometry.product_manifold import ProductManifold, ProductRiemannianMetric
from geomstats.geometry.scalar_product_metric import ScalarProductMetric


class ProductPositiveRealsAndComplexPoincareDisks(ProductManifold):
    """Class for the ProductPositiveRealsAndComplexPoincareDisks manifold.

    The positive reals and complex Poincaré disks product manifold is a
    direct product of the positive reals manifold and (n-1) complex Poincaré disks.
    Each manifold of the product is a one-dimensional manifold.

    Parameters
    ----------
    n_manifolds : int
        Number of manifolds of the product.
    """

    def __init__(self, n_manifolds, equip=True):
        self.n_manifolds = n_manifolds

        factors = [PositiveReals()] + [
            ComplexPoincareDisk() for _ in range(n_manifolds - 1)
        ]

        scales = [float(n_manifolds - i_manifold) for i_manifold in range(n_manifolds)]
        factors[0].equip_with_metric(ScalarProductMetric(factors[0], scales[0]))
        for factor, scale in zip(factors[1:], scales[1:]):
            factor.equip_with_metric(ScalarProductMetric(factor, scale))

        super().__init__(factors=factors, point_ndim=3, equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return ProductPositiveRealsAndComplexPoincareDisksMetric


class ProductPositiveRealsAndComplexPoincareDisksMetric(ProductRiemannianMetric):
    """Class defining the ProductPositiveRealsAndComplexPoincareDisks metric.

    The positive reals and complex Poincaré disks product metric is a product
    of the positive reals metric and (n-1) complex Poincaré metrics, each of them
    being multiplied by a specific constant factor (see [Cabanes_2022]_,
    [Cabanes_CESAR_2019]_, [Cabanes_RADAR_2019]_, [Cabanes_GSI_2019]_,
    [Le_Brigant_2017], [Jeuris_2016]_ and [Yang_2013]_).
    This metric comes from a model used to represent one-dimensional complex
    stationary centered Gaussian autoregressive times series.
    Such a times series can be seen as a realization of
    a multidimensional complex Gaussian distributions with zero mean,
    a Toeplitz HPD covariance matrix and a zero relation matrix.
    The ProductPositiveRealsAndComplexPoincareDisks metric is inspired by
    information geometry on this specific set of Gaussian distributions.

    References
    ----------
    .. [Cabanes_2022] Yann Cabanes. Multidimensional complex stationary
        centered Gaussian autoregressive time series machine learning
        in Poincaré and Siegel disks: application for audio and radar
        clutter classification, PhD thesis, tel-03708515, 2022.
        https://theses.hal.science/tel-03708515
    .. [Cabanes_CESAR_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon et
        Jérémie Bigot. Unsupervised Machine Learning for Pathological Radar Clutter
        Clustering: the P-Mean-Shift Algorithm, IEEE, C&ESAR 2019, Rennes, France, 2019.
        https://hal.archives-ouvertes.fr/hal-02875430
    .. [Cabanes_RADAR_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon et
        Jérémie Bigot. Non-Supervised High Resolution Doppler Machine Learning for
        Pathological Radar Clutter, IEEE, RADAR 2019, Toulon, France, 2019.
        https://hal.archives-ouvertes.fr/hal-02875415
    .. [Cabanes_GSI_2019] Yann Cabanes, Frédéric Barbaresco, Marc Arnaudon et
        Jérémie Bigot. Toeplitz Hermitian Positive Definite Matrix Machine Learning
        based on Fisher Metric, IEEE, GSI 2019, Toulouse, France, 2019.
        https://hal.archives-ouvertes.fr/hal-02875403
    .. [Le_Brigant_2017] Alice Le Brigant. Probability on the spaces of curves and
        the associated metric spaces using information geometry; radar applications,
        PhD thesis, tel-01635258, 2017.
        https://theses.hal.science/tel-01635258
    .. [Jeuris_2016] B. Jeuris and R. Vandebril. The Kahler mean of Block-Toeplitz
        matrices with Toeplitz structured blocks, 2016.
        https://epubs.siam.org/doi/pdf/10.1137/15M102112X
    .. [Yang_2013] Marc Arnaudon, Frédéric Barbaresco and Le Yang. Riemannian Medians
        and Means With Applications to Radar Signal Processing, IEEE, 2013.
    """



import geomstats.backend as gs
from geomstats.test_cases.geometry.base import VectorSpaceOpenSetTestCase
from geomstats.test_cases.information_geometry.base import (
    InformationManifoldMixinTestCase,
)


class BinomialDistributionsTestCase(
    InformationManifoldMixinTestCase, VectorSpaceOpenSetTestCase
):
    def _check_sample_belongs_to_support(self, sample, _atol):
        support = set(range(self.space.n_draws + 1))
        empty_set = set({int(val) for val in gs.unique(sample)}).difference(support)
        self.assertTrue(len(empty_set) == 0)



"""Checks and associated errors."""

import math

import geomstats.backend as gs


def check_integer(n, n_name):
    """Raise an error if n is not a > 0 integer.

    Parameters
    ----------
    n : unspecified
       Parameter to be tested.
    n_name : string
       Name of the parameter.
    """
    if n is not None and not (isinstance(n, int) and n > 0) and n != math.inf:
        raise ValueError(
            f"{n_name} is required to be either"
            " None, math.inf or a strictly positive integer,"
            f" got {n}."
        )


def check_positive(param, param_name):
    """Raise an error if param is not a > 0 number.

    Parameters
    ----------
    param : unspecified
       Parameter to be tested.
    param_name : string
       Name of the parameter.
    """
    if not (
        (isinstance(param, (int, float)) or (gs.is_array(param) and param.ndim == 0))
        and param > 0
    ):
        raise ValueError(f"{param_name} must be positive.")


def check_belongs(point, manifold, atol=gs.atol):
    """Raise an error if point does not belong to the input manifold.

    Parameters
    ----------
    point : array-like
        Point to be tested.
    manifold : Manifold
        Manifold to which the point should belong.
    manifold_name : string
        Name of the manifold for the error message.
    """
    if not gs.all(manifold.belongs(point, atol=atol)):
        raise RuntimeError(
            f"Some points do not belong to manifold '{type(manifold).__name__}'"
            f" of dimension {manifold.dim}."
        )


def check_parameter_accepted_values(param, param_name, accepted_values):
    """Raise an error if parameter does not belong to a set of values.

    Parameters
    ----------
    param : unspecified
        Parameter to be tested.
    param_name : string
        Name of the parameter.
    accepted_values : list
        Accepted values that the parameter can take.
    """
    if param not in accepted_values:
        raise ValueError(
            f"Parameter {param_name} needs to be in {accepted_values}, got: {param}."
        )


def check_point_shape(point, manifold, suppress_error=False):
    """Check if the shape of point does not match the shape of a manifold or metric.

    If the final elements of the shape of point do not match the shape of manifold
    (which may be any object with a shape attribute, such as a Riemannian metric) then
    point cannot be an array of points on the manifold (or similar) and a ValueError is
    raised. The error can be suppressed by setting suppress_error to True.

    Parameters
    ----------
    point : array-like
        The point to check the shape of.
    manifold : {Manifold, RiemannianMetric}
        The object to check the point against
    suppress_error : bool
        Whether to suppress the ShapeError if the shapes do not match. Optional, default
        is False.

    Returns
    -------
    shapes_match : bool
        Whether the shape of the point matches the shape of the manifold or metric.

    Raises
    ------
    ValueError
        If the final dimensions of point are not equal to the final dimensions of
        manifold.
    """
    representation_type = -1 * len(manifold.shape)
    shapes_match = (
        point.shape[representation_type:] == manifold.shape[representation_type:]
    )
    if not suppress_error and not shapes_match:
        shape_error_msg = (
            f"The shape of {point}, which is {point.shape}, is not"
            f" compatible with the shape of the {type(manifold).__name__}"
            f" object, which is {manifold.shape}."
        )
        raise ShapeError(shape_error_msg)
    return shapes_match


class ShapeError(ValueError):
    """Raised when there is an incompatibility between shapes."""



import numpy as _np

from geomstats._backend._dtype_utils import (
    _dyn_update_dtype,
    _modify_func_default_dtype,
    get_default_cdtype,
    get_default_dtype,
)

from .._shared_numpy._common import (
    _add_default_dtype_by_casting,
    _allow_complex_dtype,
    _box_binary_scalar,
    _box_unary_scalar,
    _cast_fout_to_input_dtype,
    _cast_out_from_dtype,
    _cast_out_to_input_dtype,
    _get_wider_dtype,
    _is_boolean,
    _is_iterable,
    as_dtype,
    atol,
    cast,
    convert_to_wider_dtype,
    is_array,
    is_bool,
    is_complex,
    is_floating,
    rtol,
    set_default_dtype,
    to_ndarray,
)

array = _cast_out_from_dtype(target=_np.array, dtype_pos=1)
eye = _modify_func_default_dtype(target=_np.eye)
zeros = _dyn_update_dtype(target=_np.zeros, dtype_pos=1)



import geomstats.backend as gs
from geomstats.test.test_case import TestCase


class SubRiemannianMetricTestCase(TestCase):
    def test_inner_coproduct(
        self, cotangent_vec_a, cotangent_vec_b, base_point, expected
    ):
        result = self.space.metric.inner_coproduct(
            cotangent_vec_a, cotangent_vec_b, base_point
        )
        self.assertAllClose(result, expected)

    def test_hamiltonian(self, cotangent_vec, base_point, expected):
        state = gs.array([base_point, cotangent_vec])
        result = self.space.metric.hamiltonian(state)
        self.assertAllClose(result, expected)

    def test_exp(self, cotangent_vec, base_point, n_steps):
        result = self.space.metric.exp(cotangent_vec, base_point, n_steps=n_steps)
        expected = base_point + cotangent_vec
        self.assertAllClose(result, expected)

    def test_geodesic(
        self,
        test_initial_point,
        test_initial_cotangent_vec,
        test_times,
        n_steps,
        expected,
    ):
        result = self.space.metric.geodesic(
            initial_point=test_initial_point,
            initial_cotangent_vec=test_initial_cotangent_vec,
            n_steps=n_steps,
        )(test_times)
        self.assertAllClose(result, expected)

    def test_symp_grad(self, test_state, expected):
        result = self.space.metric.symp_grad(hamiltonian=self.space.metric.hamiltonian)(
            test_state
        )
        self.assertAllClose(result, expected)



"""Class for the Sasaki metric.

A class implementing the Sasaki metric: The natural metric on the tangent
bundle TM of a Riemannian manifold M.

Lead authors: E. Nava-Yazdani, F. Ambellan, M. Hanik and C. von Tycowicz.
"""

from joblib import Parallel, delayed

import geomstats.backend as gs
from geomstats.geometry.base import Manifold
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import check_is_batch


class GradientDescent:
    """Gradient descent algorithm."""

    def __init__(self, lrate=0.1, max_iter=100, tol=1e-6):
        self.lrate = lrate
        self.max_iter = max_iter
        self.tol = tol

    def minimize(self, x_ini, i_pt, e_pt, grad, exp):
        """Apply a gradient descent until max_iter or a given tolerance is reached."""
        x = x_ini
        for _ in range(self.max_iter):
            grad_x = grad(x, i_pt, e_pt)
            grad_norm = gs.linalg.norm(grad_x)
            if grad_norm < self.tol:
                break
            grad_x = -self.lrate * grad_x
            x = exp(grad_x, x)
        return x


class TangentBundle(Manifold):
    """Tangent bundle of a space."""

    def __init__(self, space, equip=True):
        self.space = space
        super().__init__(dim=2 * space.dim, shape=(2,) + space.shape, equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return SasakiMetric

    def _unstack(self, point):
        return (
            point[..., 0, -self.space.point_ndim + 1 :],
            point[..., 1, -self.space.point_ndim + 1 :],
        )

    def _stack(self, space_point, space_tangent_vec):
        return gs.stack([space_point, space_tangent_vec], axis=-self.point_ndim)

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Point to evaluate.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the manifold.
        """
        space_point, space_tangent_vec = self._unstack(point)
        return gs.logical_and(
            self.space.belongs(space_point),
            self.space.is_tangent(space_tangent_vec, space_point),
        )

    def random_point(self, n_samples=1, bound=1.0):
        """Sample random points on the manifold according to some distribution.

        If the manifold is compact, preferably a uniform distribution will be used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound of the interval in which to sample for non compact manifolds.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., *point_shape]
            Points sampled on the manifold.
        """
        space_point = self.space.random_point(n_samples, bound)
        space_tangent_vec = self.space.random_tangent_vec(space_point)
        return self._stack(space_point, space_tangent_vec)

    @staticmethod
    def projection(point):
        """Project a point to the vector space.

        This method is for compatibility and returns `point`. `point` should
        have the right shape,

        Parameters
        ----------
        point: array-like, shape[..., *point_shape]
            Point.

        Returns
        -------
        point: array-like, shape[..., *point_shape]
            Point.
        """
        return gs.copy(point)

    def is_tangent(self, vector, base_point=None, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Tangent vectors are identified with points of the vector space so
        this checks the shape of the input vector.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point in the vector space.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : array-like, shape=[...,]
            Boolean denoting if vector is a tangent vector at the base point.
        """
        raise NotImplementedError("`is_tangent` is not implemented")

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space.

        This method is for compatibility and returns vector.

        Parameters
        ----------
        vector : array-like, shape=[..., *point_shape]
            Vector.
        base_point : array-like, shape=[..., *point_shape]
            Point in the vector space

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        raise NotImplementedError("`to_tangent` is not implemented")

    def random_tangent_vec(self, base_point, n_samples=1):
        """Generate random tangent vec."""
        raise NotImplementedError("`random_tangent_vec` is not implemented")


class SasakiMetric(RiemannianMetric):
    """Implements of the Sasaki metric on the tangent bundle TM of a Riem. manifold M.

    The Sasaki metric is characterized by the following three properties:

    * the canonical projection of TM becomes a Riemannian submersion,
    * parallel vector fields along curves are orthogonal to their fibres, and
    * its restriction to any tangent space is Euclidean.

    Geodesic computations are realized via a discrete formulation of the
    geodesic equation on TM that involve geodesics, parallel translation, and
    the curvature tensor on the base manifold M (see [1]_ for details).
    However, as the implemented energy in the discrete-geodesics-optimization
    as well as the approximations of its gradient slightly differ from those
    proposed in [1]_, we also refer to [2]_ for additional details.

    Parameters
    ----------
    space : Manifold
        Tangent bundle.
    n_jobs: int
        Number of jobs for parallel computing.
        Optional, default: 1.
    n_steps : int
        Number of discrete time steps.
        Optional, default: 3.

    References
    ----------
    .. [1] Muralidharan, P., & Fletcher, P. T. "Sasaki metrics for analysis of
        longitudinal data on manifolds", IEEE CVPR 2012, pp. 1027-1034
        https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4270017/
    .. [2] Nava-Yazdani, E., Hanik, M., Ambellan, F., & von Tycowicz, C.
        "On Gradient Formulas in an Algorithm for the Logarithm of the Sasaki
        Metric", Technical Report Zuse-Institut Berlin, 2022
        https://nbn-resolving.org/urn/resolver.pl?urn:nbn:de:0297-zib-87174
    """

    def __init__(self, space, n_jobs=1, n_steps=3):
        super().__init__(space=space)
        self.n_jobs = n_jobs
        self.n_steps = n_steps
        self._gradient_descent = GradientDescent()

    def exp(self, tangent_vec, base_point):
        """Compute the Riemannian exponential of a point.

        Exponential map at base_point of tangent_vec computed by
        shooting a Sasaki geodesic using an Euler integration on TM.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 2, M.dim]
            Tangent vector in TTM at the base point in TM.
        base_point : array-like, shape=[..., 2, M.dim]
            Point in the tangent bundle TM of manifold M.


        Returns
        -------
        exp : array-like, shape=[..., 2, M.dim]
            Point on the tangent bundle TM.
        """
        par_trans = self._space.space.metric.parallel_transport
        eps = 1 / self.n_steps

        v0, w0 = self._space._unstack(tangent_vec)
        p0, u0 = self._space._unstack(base_point)

        for _ in range(self.n_steps):
            p = self._space.space.metric.exp(eps * v0, p0)
            u = par_trans(u0 + eps * w0, p0, end_point=p)
            v = par_trans(
                v0 - eps * (self._space.space.metric.curvature(u0, w0, v0, p0)),
                p0,
                end_point=p,
            )
            w = par_trans(w0, p0, end_point=p)
            p0, u0 = p, u
            v0, w0 = v, w

        return self._space._stack(p, u)

    def log(self, point, base_point):
        """Compute the Riemannian logarithm of a point.

        Logarithmic map at base_point of point computed by iteratively relaxing
        a discretized geodesic between base_point and point.

        Parameters
        ----------
        point : array-like, shape=[..., 2, M.dim]
            Point in the tangent bundle TM of manifold M.
        base_point : array-like, shape=[..., 2, M.dim]
            Point in the tangent bundle TM of manifold M.

        Returns
        -------
        log : array-like, shape=[..., 2, M.dim]
            Tangent vector at the base point equal to the Riemannian logarithm
            of point at the base point.
        """
        par_trans = self._space.space.metric.parallel_transport

        pu = self.geodesic_discrete(base_point, point)
        pu1 = gs.take(pu, 1, axis=-(self._space.point_ndim + 1))

        p1, u1 = self._space._unstack(pu1)
        p0, u0 = self._space._unstack(base_point)

        w = par_trans(u1, p1, end_point=p0) - u0
        v = self._space.space.metric.log(point=p1, base_point=p0)
        return self.n_steps * self._space._stack(v, w)

    def geodesic_discrete(self, initial_point, end_point):
        """Compute Sakai geodesic employing a variational time discretization.

        Parameters
        ----------
        end_points : array-like, shape=[..., 2, M.shape]
            Points in the tangent bundle TM of manifold M.
        initial_points : array-like, shape=[..., 2, M.shape]
            Points in the tangent bundle TM of manifold M.

        Returns
        -------
        geodesic : array-like, shape=[..., n_steps + 1, 2, M.shape]
            Discrete geodesics of form x(s)=(p(s), u(s)) in Sasaki metric
            connecting initial_point = x(0) and end_point = x(1).
        """
        metric = self._space.space.metric
        par_trans = metric.parallel_transport

        def _grad(pu, i_pt, e_pt):
            """Gradient of discrete geodesic energy."""
            pu = gs.vstack(
                [
                    gs.expand_dims(i_pt, axis=0),
                    pu,
                    gs.expand_dims(e_pt, axis=0),
                ]
            )
            p, u = self._space._unstack(pu)

            p1, p2, p3 = p[:-2], p[1:-1], p[2:]
            u1, u2, u3 = u[:-2], u[1:-1], u[2:]

            eps = 1 / self.n_steps

            v2 = metric.log(p3, p2) / eps
            w2 = (par_trans(u3, p3, end_point=p2) - u2) / eps

            gp = (metric.log(p3, p2) + metric.log(p1, p2)) / (
                2 * eps**2
            ) - metric.curvature(u2, w2, v2, p2)

            gu = (
                par_trans(u3, p3, end_point=p2)
                - 2 * u2
                + par_trans(u1, p1, end_point=p2)
            ) / eps**2

            return -self._space._stack(gp, gu) * eps

        def _geodesic_discrete_single(initial_point, end_point):
            """Calculate the discrete geodesic."""
            ijk = "ijk"[: self._space.space.point_ndim]

            def _scalarmul(scalar, point):
                return gs.einsum(f"...,...{ijk}->...{ijk}", scalar, point)

            p0, u0 = initial_point[0], initial_point[1]
            pL, uL = end_point[0], end_point[1]

            v = metric.log(pL, p0)
            s = gs.linspace(0.0, 1.0, self.n_steps + 1)[1:-1]

            p_ini = metric.exp(gs.einsum(f"p, {ijk}->p{ijk}", s, v), p0)
            u_ini = _scalarmul(
                (1.0 - s), par_trans(u0, p0, end_point=p_ini)
            ) + _scalarmul(s, par_trans(uL, pL, end_point=p_ini))
            pu_ini = self._space._stack(p_ini, u_ini)

            x = self._gradient_descent.minimize(
                pu_ini, initial_point, end_point, _grad, self.exp
            )

            return gs.vstack(
                [
                    gs.expand_dims(initial_point, axis=0),
                    x,
                    gs.expand_dims(end_point, axis=0),
                ]
            )

        is_batch = check_is_batch(self._space.point_ndim, initial_point, end_point)
        if not is_batch:
            return _geodesic_discrete_single(initial_point, end_point)

        initial_point, end_point = gs.broadcast_arrays(initial_point, end_point)

        with Parallel(n_jobs=min(self.n_jobs, len(end_point)), verbose=0) as parallel:
            rslt = parallel(
                delayed(_geodesic_discrete_single)(i_pt, e_pt)
                for i_pt, e_pt in zip(initial_point, end_point)
            )
        return gs.array(rslt)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., 2, M.dim]
            Tangent vector in TTM of the tangent bundle TM.
        tangent_vec_b : array-like, shape=[..., 2, M.dim]
            Tangent vector in TTM of the tangent bundle TM.
        base_point : array-like, shape=[..., 2, M.dim]
            Point in the tangent bundle TM of manifold M.

        Returns
        -------
        inner_product : array-like, shape=[..., 1]
            Inner-product.
        """
        vec_a_0, vec_a_1 = self._space._unstack(tangent_vec_a)
        vec_b_0, vec_b_1 = self._space._unstack(tangent_vec_b)
        pt, _ = self._space._unstack(base_point)

        inner = self._space.space.metric.inner_product
        return inner(vec_a_0, vec_b_0, pt) + inner(vec_a_1, vec_b_1, pt)



import pytest

import geomstats.backend as gs
from geomstats.geometry.symmetric_matrices import SymmetricMatrices
from geomstats.test_cases.information_geometry.dirichlet import (
    DirichletDistributionsTestCase,
    DirichletMetricTestCase,
)

#  TODO: add againts numpy test case for pdf?


def sectional_curvature(tangent_vec_a, tangent_vec_b, base_point):
    x, y = base_point[..., 0], base_point[..., 1]
    detg = gs.polygamma(1, x) * gs.polygamma(1, y) - gs.polygamma(1, x + y) * (
        gs.polygamma(1, x) + gs.polygamma(1, y)
    )
    return (
        gs.polygamma(2, x)
        * gs.polygamma(2, y)
        * gs.polygamma(2, x + y)
        * (
            gs.polygamma(1, x) / gs.polygamma(2, x)
            + gs.polygamma(1, y) / gs.polygamma(2, y)
            - gs.polygamma(1, x + y) / gs.polygamma(2, x + y)
        )
        / (4 * detg**2)
    )


def metric_matrix(base_point):
    param_a = base_point[..., 0]
    param_b = base_point[..., 1]
    vector = gs.stack(
        [
            gs.polygamma(1, param_a) - gs.polygamma(1, param_a + param_b),
            -gs.polygamma(1, param_a + param_b),
            gs.polygamma(1, param_b) - gs.polygamma(1, param_a + param_b),
        ],
        axis=-1,
    )
    return SymmetricMatrices.matrix_representation(vector)


def christoffels(base_point):
    def coefficients(param_a, param_b):
        """Christoffel coefficients for the beta distributions."""
        poly1a = gs.polygamma(1, param_a)
        poly2a = gs.polygamma(2, param_a)
        poly1b = gs.polygamma(1, param_b)
        poly2b = gs.polygamma(2, param_b)
        poly1ab = gs.polygamma(1, param_a + param_b)
        poly2ab = gs.polygamma(2, param_a + param_b)
        metric_det = 2 * (poly1a * poly1b - poly1ab * (poly1a + poly1b))

        c1 = (poly2a * (poly1b - poly1ab) - poly1b * poly2ab) / metric_det
        c2 = -poly1b * poly2ab / metric_det
        c3 = (poly2b * poly1ab - poly1b * poly2ab) / metric_det
        return c1, c2, c3

    param_a, param_b = base_point[..., 0], base_point[..., 1]
    c1, c2, c3 = coefficients(param_a, param_b)
    c4, c5, c6 = coefficients(param_b, param_a)
    vector_0 = gs.stack([c1, c2, c3], axis=-1)
    vector_1 = gs.stack([c6, c5, c4], axis=-1)
    gamma_0 = SymmetricMatrices.matrix_representation(vector_0)
    gamma_1 = SymmetricMatrices.matrix_representation(vector_1)

    return gs.stack([gamma_0, gamma_1], axis=-3)


class BetaDistributionsTestCase(DirichletDistributionsTestCase):
    def _check_sample_belongs_to_support(self, sample, atol):
        self.assertTrue(gs.all(sample >= 0))
        self.assertTrue(gs.all(sample <= 1))


class BetaMetricTestCase(DirichletMetricTestCase):
    def test_metric_det(self, point, expected, atol):
        res = self.space.metric.metric_det(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_metric_det_against_metric_matrix(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        res = self.space.metric.metric_det(point)
        expected = gs.linalg.det(self.space.metric.metric_matrix(point))
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_metric_det_lower_bound(self, n_points):
        """Check metric determinant lower bound.

        References
        ----------
        .. [BP2019] Brigant, A.L. and Puechmorel, S. (2019)
        ‘The Fisher-Rao geometry of beta distributions applied to the study of
        canonical moments’. arXiv.
        Available at: https://doi.org/10.48550/arXiv.1904.08247.
        """
        base_point = self.data_generator.random_point(n_points)
        res = self.space.metric.metric_det(base_point)

        alpha = base_point[..., 0]
        beta = base_point[..., 1]

        num = 1 + alpha + beta
        den = 2 * alpha * beta * gs.power((alpha + beta), 2)
        lower_bound = num / den
        self.assertTrue(gs.all(res > lower_bound))

    @pytest.mark.random
    def test_sectional_curvature_against_closed_form(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        res = self.space.metric.sectional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        expected = sectional_curvature(tangent_vec_a, tangent_vec_b, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_sectional_curvature_lower_bound(self, n_points):
        """Check sectional curvature lower bound.

        It is still a conjecture (see [BPP2020]_).

        References
        ----------
        .. [BPP2020] Brigant, A.L., Preston, S. and Puechmorel, S. (2021)
        ‘Fisher-Rao geometry of Dirichlet distributions’, Differential Geometry
        and its Applications, 74, p. 101702.
        Available at: https://doi.org/10.1016/j.difgeo.2020.101702.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        res = self.space.metric.sectional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        lower_bound = -1 / 2
        self.assertTrue(gs.all(res > lower_bound))

    @pytest.mark.random
    def test_metric_matrix_against_closed_form(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        res = self.space.metric.metric_matrix(base_point)
        expected = metric_matrix(base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_christoffels_against_closed_form(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        res = self.space.metric.christoffels(base_point)
        expected = christoffels(base_point)
        self.assertAllClose(res, expected, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import LevelSetTestCase
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase
from geomstats.test_cases.information_geometry.base import (
    InformationManifoldMixinTestCase,
)


def fisher_rao_dist(point_a, point_b, n_draws):
    return 2 * gs.sqrt(n_draws) * gs.arccos(gs.sum(gs.sqrt(point_a * point_b), axis=-1))


class MultinomialDistributionsTestCase(
    InformationManifoldMixinTestCase, LevelSetTestCase
):
    def _check_sample_belongs_to_support(self, sample, _atol):
        self.assertTrue(gs.all(sample >= 0.0))
        self.assertTrue(gs.all(gs.sum(sample, axis=-1) == self.space.n_draws))


class MultinomialMetricTestCase(RiemannianMetricTestCase):
    def test_simplex_to_sphere(self, point, expected, atol):
        res = self.space.metric.simplex_to_sphere(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_simplex_to_sphere_belongs(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        point_sphere = self.space.metric.simplex_to_sphere(point)

        res = self.space.metric._sphere.belongs(point_sphere, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    def test_sphere_to_simplex(self, point, expected, atol):
        res = self.space.metric.sphere_to_simplex(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_sphere_to_simplex_vec(self, n_reps, atol):
        point = self.data_generator_sphere.random_point()

        expected = self.space.metric.sphere_to_simplex(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_sphere_to_simplex_belongs(self, n_points, atol):
        point = self.data_generator_sphere.random_point(n_points)

        point_simplex = self.space.metric.sphere_to_simplex(point)
        expected = gs.ones(n_points, dtype=bool)

        res = self.space.belongs(point_simplex, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_sphere_to_simplex_after_simplex_to_sphere(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        point_sphere = self.space.metric.simplex_to_sphere(point)
        point_ = self.space.metric.sphere_to_simplex(point_sphere)

        self.assertAllClose(point_, point, atol=atol)

    @pytest.mark.random
    def test_simplex_to_sphere_after_sphere_to_simplex(self, n_points, atol):
        point_sphere = self.data_generator_sphere.random_point(n_points)

        point = self.space.metric.sphere_to_simplex(point_sphere)
        point_sphere_ = self.space.metric.simplex_to_sphere(point)

        self.assertAllClose(point_sphere_, gs.abs(point_sphere), atol=atol)

    def test_tangent_simplex_to_sphere(self, tangent_vec, base_point, expected, atol):
        res = self.space.metric.tangent_simplex_to_sphere(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_tangent_simplex_to_sphere_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        tangent_vec_sphere = self.space.metric.tangent_simplex_to_sphere(
            tangent_vec, base_point
        )
        base_point_sphere = self.space.metric.simplex_to_sphere(base_point)

        is_tangent = self.space.metric._sphere.is_tangent(
            tangent_vec_sphere, base_point_sphere, atol=atol
        )
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(is_tangent, expected)

    def test_tangent_sphere_to_simplex(self, tangent_vec, base_point, expected, atol):
        res = self.space.metric.tangent_sphere_to_simplex(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tangent_sphere_to_simplex_vec(self, n_reps, atol):
        base_point = self.data_generator_sphere.random_point()
        tangent_vec = self.data_generator_sphere.random_tangent_vec(base_point)

        expected = self.space.metric.tangent_sphere_to_simplex(tangent_vec, base_point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tangent_sphere_to_simplex_is_tangent(self, n_points, atol):
        base_point_sphere = self.data_generator_sphere.random_point(n_points)
        tangent_vec_sphere = self.data_generator_sphere.random_tangent_vec(
            base_point_sphere
        )

        tangent_vec = self.space.metric.tangent_sphere_to_simplex(
            tangent_vec_sphere, base_point_sphere
        )
        base_point = self.space.metric.sphere_to_simplex(base_point_sphere)

        is_tangent = self.space.is_tangent(tangent_vec, base_point, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(is_tangent, expected)

    @pytest.mark.random
    def test_tangent_sphere_to_simplex_after_tangent_simplex_to_sphere(
        self, n_points, atol
    ):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        tangent_vec_sphere = self.space.metric.tangent_simplex_to_sphere(
            tangent_vec, base_point
        )
        base_point_sphere = self.space.metric.simplex_to_sphere(base_point)

        tangent_vec_ = self.space.metric.tangent_sphere_to_simplex(
            tangent_vec_sphere, base_point_sphere
        )
        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    @pytest.mark.random
    def test_tangent_simplex_to_sphere_after_tangent_sphere_to_simplex(
        self, n_points, atol
    ):
        base_point_sphere = gs.abs(self.data_generator_sphere.random_point(n_points))
        tangent_vec_sphere = self.data_generator_sphere.random_tangent_vec(
            base_point_sphere
        )

        tangent_vec = self.space.metric.tangent_sphere_to_simplex(
            tangent_vec_sphere, base_point_sphere
        )
        base_point = self.space.metric.sphere_to_simplex(base_point_sphere)

        tangent_vec_sphere_ = self.space.metric.tangent_simplex_to_sphere(
            tangent_vec, base_point
        )

        self.assertAllClose(tangent_vec_sphere_, tangent_vec_sphere, atol=atol)

    @pytest.mark.random
    def test_sectional_curvature_is_positive(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        res = self.space.metric.sectional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertTrue(gs.all(res > 0))

    @pytest.mark.random
    def test_dist_against_closed_form(self, n_points, atol):
        point_a = self.data_generator.random_point(n_points)
        point_b = self.data_generator.random_point(n_points)

        res = self.space.metric.dist(point_a, point_b)
        expected = fisher_rao_dist(point_a, point_b, self.space.n_draws)
        self.assertAllClose(res, expected, atol=atol)



import geomstats.backend as gs
from geomstats.test.test_case import TestCase


class HyperbolicEmbeddingTestCase(TestCase):
    def test_log_sigmoid(self, point, expected, atol):
        res = self.embedding.log_sigmoid(point)
        self.assertAllClose(res, expected, atol=atol)

    def test_grad_log_sigmoid(self, point, expected, atol):
        res = self.embedding.grad_log_sigmoid(point)
        self.assertAllClose(res, expected, atol=atol)

    def test_loss(
        self, point, point_context, point_negative, expected_loss, expected_grad, atol
    ):
        loss_value, loss_grad = self.embedding.loss(
            point, point_context, point_negative
        )
        self.assertAllClose(loss_value, expected_loss, atol=atol)
        self.assertAllClose(loss_grad, expected_grad, atol=atol)

    def test_embed(self, graph, atol):
        embeddings = self.embedding.embed(graph)
        res = self.embedding.manifold.belongs(embeddings, atol)
        self.assertAllEqual(res, gs.ones_like(res))



"""The n-dimensional hyperbolic space.

Poincare half-space representation.

Lead author: Alice Le Brigant.
"""

import math

import geomstats.backend as gs
from geomstats.geometry._hyperbolic import HyperbolicDiffeo, _Hyperbolic
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.poincare_ball import PoincareBall
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.vectorization import repeat_out


class PoincareHalfSpace(_Hyperbolic, VectorSpaceOpenSet):
    """Class for the n-dimensional Poincare half-space.

    Class for the n-dimensional Poincaré half space model. For other
    representations of hyperbolic spaces see the `Hyperbolic` class.

    Parameters
    ----------
    dim : int
        Dimension of the hyperbolic space.
    """

    def __init__(self, dim, equip=True):
        self.coords_type = "half-space"
        super().__init__(
            dim=dim,
            embedding_space=Euclidean(dim),
            intrinsic=True,
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return PoincareHalfSpaceMetric

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the upper half space.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point to be checked.
        atol : float
            Absolute tolerance to evaluate positivity of the last coordinate.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Array of booleans indicating whether the corresponding
            points belong to the upper half space.
        """
        point_dim = point.shape[-1]
        belongs = point_dim == self.dim
        return gs.logical_and(belongs, point[..., -1] >= -atol)

    def projection(self, point, atol=gs.atol):
        """Project a point in ambient space to the open set.

        The last coordinate is floored to `atol` if it is negative.

        Parameters
        ----------
        point : array-like, shape=[..., dim_embedding]
            Point in ambient space.
        atol : float
            Tolerance to evaluate positivity.

        Returns
        -------
        projected : array-like, shape=[..., dim_embedding]
            Projected point.
        """
        last = gs.where(point[..., -1] < atol, atol, point[..., -1])
        return gs.concatenate([point[..., :-1], last[..., None]], axis=-1)


class PoincareHalfSpaceMetric(PullbackDiffeoMetric):
    """Class for the metric of the n-dimensional hyperbolic space.

    Class for the metric of the n-dimensional hyperbolic space
    as embedded in the Poincaré half space model.
    """

    def __init__(self, space):
        image_space = PoincareBall(dim=space.dim)
        super().__init__(
            space=space,
            image_space=image_space,
            diffeo=HyperbolicDiffeo(space.coords_type, image_space.coords_type),
        )

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Compute the inner-product of two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., dim + 1]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., dim + 1]
            Second tangent vector at base point.
        base_point : array-like, shape=[..., dim + 1]
            Point in hyperbolic space.

        Returns
        -------
        inner_prod : array-like, shape=[..., 1]
            Inner-product of the two tangent vectors.
        """
        inner_prod = gs.sum(tangent_vec_a * tangent_vec_b, axis=-1)
        return inner_prod / base_point[..., -1] ** 2

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base
        point onto its image.
        In the case of the hyperbolic space, it does not depend on the base
        point and is infinite everywhere, because of the negative curvature.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)



"""Information Manifold of multivariate normal distributions with the Fisher metric.

Lead authors: Antoine Collas, Alice Le Brigant.
"""

import math

from scipy.stats import multivariate_normal, norm

import geomstats.backend as gs
import geomstats.errors as errors
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.diffeo import Diffeo
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.poincare_half_space import PoincareHalfSpace
from geomstats.geometry.product_manifold import ProductManifold
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.geometry.scalar_product_metric import ScalarProductMetric
from geomstats.geometry.spd_matrices import SPDAffineMetric, SPDMatrices
from geomstats.information_geometry.base import (
    InformationManifoldMixin,
    ScipyMultivariateRandomVariable,
    ScipyUnivariateRandomVariable,
)
from geomstats.vectorization import broadcast_to_multibatch, get_batch_shape, repeat_out


class NormalDistributions:
    """Class for the normal distributions.

    This class is a common interface to the following different situations:

    - univariate normal distributions
    - centered multivariate normal distributions
    - multivariate normal distributions with diagonal covariance matrix
    - general multivariate normal distributions

    Parameters
    ----------
    sample_dim : int
        Dimension of the sample space of the normal distribution.
    distribution_type : str, {'centered', 'diagonal', 'general'}
        Type of distributions.
        Optional, default: 'general'.
    """

    def __new__(cls, sample_dim, distribution_type="general", equip=True):
        """Instantiate class that corresponds to the distribution_type."""
        errors.check_parameter_accepted_values(
            distribution_type,
            "distribution_type",
            ["centered", "diagonal", "general"],
        )
        if sample_dim == 1:
            return UnivariateNormalDistributions(equip=equip)
        if distribution_type == "centered":
            return CenteredNormalDistributions(sample_dim, equip=equip)
        if distribution_type == "diagonal":
            return DiagonalNormalDistributions(sample_dim, equip=equip)
        return GeneralNormalDistributions(sample_dim, equip=equip)


class UnivariateNormalDistributions(InformationManifoldMixin, PoincareHalfSpace):
    """Class for the manifold of univariate normal distributions.

    This is upper half-plane.
    """

    def __init__(self, equip=True):
        super().__init__(dim=2, support_shape=(), equip=equip)
        self._scp_rv = UnivariateNormalDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return UnivariateNormalMetric

    @staticmethod
    def random_point(n_samples=1, bound=1.0):
        """Sample parameters of normal distributions.

        The uniform distribution on [-bound/2, bound/2]x[0, bound] is used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Side of the square where the normal parameters are sampled.
            Optional, default: 5.

        Returns
        -------
        samples : array-like, shape=[..., 2]
            Sample of points representing normal distributions.
        """
        means = -bound + 2 * bound * gs.random.rand(n_samples)
        stds = bound * gs.random.rand(n_samples)
        if n_samples == 1:
            return gs.array((means[0], stds[0]))
        return gs.transpose(gs.stack((means, stds)))

    def sample(self, point, n_samples=1):
        """Sample from the normal distribution.

        Sample from the normal distribution with parameters provided
        by point.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point representing a normal distribution (mean and scale).
        n_samples : int
            Number of points to sample with each pair of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples]
            Sample from normal distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pdf associated to point.

        Compute the probability density function of the normal
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., 2]
            Point representing a normal distribution (mean and scale).

        Returns
        -------
        pdf : function
            Probability density function of the normal distribution with
            parameters provided by point.
        """
        means = gs.expand_dims(point[..., 0], axis=-1)
        stds = gs.expand_dims(point[..., 1], axis=-1)
        pdf_normalization = 1.0 / gs.sqrt(2 * gs.pi * stds**2)

        def pdf(x):
            """Generate parameterized function for normal pdf.

            Parameters
            ----------
            x : array-like, shape=[n_samples,]
                Points at which to compute the probability density function.

            Returns
            -------
            pdf_at_x : array-like, shape=[..., n_samples]
                Values of pdf at x for each value of the parameters provided
                by point.
            """
            x = gs.reshape(gs.array(x), (-1,))
            return pdf_normalization * gs.exp(-((x - means) ** 2) / (2 * stds**2))

        return pdf


class CenteredNormalDistributions(InformationManifoldMixin, SPDMatrices):
    r"""Class for the manifold of centered multivariate normal distributions.

    This is the class for multivariate normal distributions with zero mean.
    Each distribution is represented by its covariance matrix, i.e. a symmetric
    positive-definite matrix of size :math:`sample\_dim`.

    Parameters
    ----------
    sample_dim : int
        Dimension of the sample space of the multivariate normal distribution.
    """

    def __init__(self, sample_dim, equip=True):
        super().__init__(n=sample_dim, support_shape=(sample_dim,), equip=equip)
        self.sample_dim = sample_dim
        self._scp_rv = SharedMeanNormalDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return CenteredNormalMetric

    def sample(self, point, n_samples=1):
        """Sample from a centered multivariate normal distribution.

        Parameters
        ----------
        point : array-like, shape=[..., sample_dim, sample_dim]
            Symmetric positive definite matrix representing the covariance matrix
            of a multivariate normal distribution with zero mean.
        n_samples : int
            Number of points to sample with each covariance matrix in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples, sample_dim]
            Sample from centered multivariate normal distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pdf associated to point.

        Parameters
        ----------
        point : array-like, shape=[..., sample_dim, sample_dim]
            Symmetric positive definite matrix representing the covariance matrix
            of a multivariate normal distribution with zero mean.

        Returns
        -------
        pdf : function
            Probability density function of the centered multivariate normal
            distributions with covariance matrices provided by point.
        """
        batch_shape = get_batch_shape(self.point_ndim, point)
        det_cov = gs.linalg.det(point)
        inv_cov = gs.linalg.inv(point)
        pdf_normalization = 1 / gs.sqrt(gs.power(2 * gs.pi, self.sample_dim) * det_cov)

        def pdf(x):
            """Generate parameterized function for normal pdf.

            Parameters
            ----------
            x : array-like, shape=[n_samples, sample_dim]
                Points at which to compute the probability
                density function.

            Returns
            -------
            pdf_at_x: array-like, shape=[..., n_samples]
                Probability density function at x.
            """
            x_, inv_cov_ = broadcast_to_multibatch(
                (x.shape[0],), batch_shape, x, inv_cov
            )

            aux = gs.exp(-0.5 * gs.dot(x_, gs.matvec(inv_cov_, x_)))
            return gs.einsum(
                "...,...i->...i", pdf_normalization, gs.to_ndarray(aux, to_ndim=1)
            )

        return pdf


class DiagonalNormalDistributions(InformationManifoldMixin, VectorSpaceOpenSet):
    r"""Class for the manifold of diagonal multivariate normal distributions.

    This is the class for multivariate normal distributions with diagonal
    covariance matrices. Each distribution is represented by a vector of size
    :math:`2 * sample\_dim` where the first :math:`sample\_dim` elements contain
    the mean vector and the :math:`sample\_dim` last elements contain the
    diagonal of the covariance matrix.

    Parameters
    ----------
    sample_dim : int
          Dimension of the sample space of the multivariate normal distribution.
    """

    def __init__(self, sample_dim, equip=True):
        self.sample_dim = sample_dim
        self.sample_space = Euclidean(dim=sample_dim, equip=False)
        dim = int(2 * sample_dim)
        super().__init__(
            dim=dim,
            embedding_space=Euclidean(dim),
            support_shape=(sample_dim,),
            equip=equip,
        )
        self._scp_rv = DiagonalNormalDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return DiagonalNormalMetric

    def _unstack_mean_diagonal(self, point):
        """Extract mean and diagonal of the covariance matrix from a given point.

        Parameters
        ----------
        sample_dim : int
            Dimension of the sample space of the multivariate normal distribution.
        point : array-like, shape=[..., 2 * sample_dim]
            Input point from which means and diagonals are extracted.

        Returns
        -------
        mean : array-like, shape=[..., sample_dim]
            Means from the input point.
        diagonal : array-like, shape=[..., sample_dim]
            Diagonals of covariance matrices from the input point.
        """
        mean = point[..., : self.sample_dim]
        diagonal = point[..., self.sample_dim :]
        return mean, diagonal

    @staticmethod
    def _stack_mean_diagonal(mean, diagonal):
        """Set mean and diagonal of the covariance matrix into a point.

        Parameters
        ----------
        mean : array-like, shape=[..., sample_dim]
            Means to stack.
        diagonal : array-like, shape=[..., sample_dim]
            Diagonals of covariance matrices from the input point.

        Returns
        -------
        point : array-like, shape=[..., 2 * sample_dim]
            Point with means and diagonals covariance matrices.
        """
        return gs.concatenate([mean, diagonal], axis=-1)

    def belongs(self, point, atol=gs.atol):
        r"""Evaluate if the point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., 2 * sample_dim]
            Point to test. First :math:`sample\_dim` elements contain the
            mean vector and the last :math:`sample\_dim` elements contain
            the diagonal of the covariance matrix.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        point_dim = point.shape[-1]
        belongs = point_dim == self.dim
        _, diagonal = self._unstack_mean_diagonal(point)
        return gs.logical_and(belongs, gs.all(diagonal >= -atol, axis=-1))

    def random_point(self, n_samples=1):
        r"""Generate random parameters of multivariate diagonal normal distributions.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., 2 * sample_dim]
            Sample of points representing multivariate diagonal normal distributions.
            First :math:`sample\_dim` elements contain the mean vector and the last
            :math:`sample\_dim` elements contain the diagonal of the covariance matrix.
        """
        sample_dim = self.sample_dim
        bound = 1.0
        mean = self.sample_space.random_point(n_samples=n_samples, bound=bound)
        if n_samples == 1:
            diagonal = gs.array(norm.rvs(size=(sample_dim,)) ** 2)
        else:
            diagonal = gs.array(norm.rvs(size=(n_samples, sample_dim)) ** 2)
        point = self._stack_mean_diagonal(mean, diagonal)
        return point

    def projection(self, point):
        r"""Project a point on the manifold.

        Floor the eigenvalues of the diagonal covariance matrix to gs.atol.

        Parameters
        ----------
        point : array-like, shape=[..., 2 * sample_dim]
            Point to project. First :math:`sample\_dim` elements contain
            the mean vector and the last :math:`sample\_dim` elements contain
            the diagonal of the covariance matrix.

        Returns
        -------
        projected: array-like, shape=[..., 2 * sample_dim]
            Point containing means and diagonals
            of covariance matrices.
        """
        mean, diagonal = self._unstack_mean_diagonal(point)
        regularized = gs.where(diagonal < gs.atol, gs.atol, diagonal)
        return self._stack_mean_diagonal(mean, regularized)

    def sample(self, point, n_samples=1):
        r"""Sample from the diagonal multivariate normal distribution.

        Parameters
        ----------
        point : array-like, shape=[..., 2 * sample_dim]
            Point on the manifold. First :math:`sample\_dim` elements contain
            the mean vector and the last :math:`sample\_dim` elements contain
            the diagonal of the covariance matrix.
        n_samples : int
            Number of points to sample with each pair of parameters in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples, sample_dim]
            Sample from multivariate normal distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        r"""Compute pdf associated to point.

        Parameters
        ----------
        point : array-like, shape=[..., 2 * sample_dim]
            Point representing a probability distribution.
            First :math:`sample\_dim` elements contain the mean vector and the last
            :math:`sample\_dim` elements contain the diagonal of the covariance matrix.

        Returns
        -------
        pdf : function
            Probability density function of the normal distribution with
            parameters provided by point.
        """
        batch_shape = get_batch_shape(self.point_ndim, point)
        n = self.sample_dim
        mean, diagonal = self._unstack_mean_diagonal(point)
        det_cov = gs.prod(diagonal, axis=-1)
        pdf_normalization = 1 / gs.sqrt(gs.power((2 * gs.pi), n) * det_cov)

        def pdf(x):
            """Generate parameterized function for normal pdf.

            Parameters
            ----------
            x : array-like, shape=[n_samples, sample_dim]
                Points at which to compute the probability
                density function.
            """
            x_, (mean_, diagonal_) = broadcast_to_multibatch(
                (x.shape[0],), batch_shape, x, mean, diagonal
            )

            aux = gs.exp(-0.5 * gs.sum(((x_ - mean_) ** 2) / diagonal_, axis=-1))
            return gs.einsum(
                "...,...i->...i", pdf_normalization, gs.to_ndarray(aux, to_ndim=1)
            )

        return pdf


class GeneralNormalDistributions(InformationManifoldMixin, ProductManifold):
    r"""Class for the manifold of multivariate normal distributions.

    This is the class for multivariate normal distributions on the Euclidean space.
    Each distribution is represented by the concatenation of its mean vector and
    its covariance matrix reshaped in a vector of size :math:`sample\_dim ** 2`.

    Parameters
    ----------
    sample_dim : int
        Dimension of the sample space of the multivariate normal distribution.
    """

    def __init__(self, sample_dim, equip=True):
        super().__init__(
            factors=(Euclidean(sample_dim), SPDMatrices(sample_dim)),
            point_ndim=1,
            support_shape=(sample_dim,),
            equip=equip,
        )
        self.sample_dim = sample_dim
        self._scp_rv = MultivariateNormalDistributionsRandomVariable(self)

    def _unstack_mean_covariance(self, point):
        """Extract mean and covariance matrix from a given point.

        Parameters
        ----------
        point : array-like, shape=[..., sample_dim + sample_dim ** 2]
            Input point from which means and covariance matrices are extracted.

        Returns
        -------
        mean : array-like, shape=[..., sample_dim]
            Means from the input point.
        diagonal : array-like, shape=[..., sample_dim, sample_dim]
            Covariance matrices from the input point.
        """
        batch_shape = get_batch_shape(self.point_ndim, point)
        mean = point[..., : self.sample_dim]
        cov = point[..., self.sample_dim :]
        cov = cov.reshape(batch_shape + (self.sample_dim, self.sample_dim))
        return mean, cov

    def sample(self, point, n_samples=1):
        r"""Sample from a multivariate normal distribution.

        Parameters
        ----------
        point : array-like, shape=[..., sample_dim + sample_dim ** 2]
            Point representing a multivariate normal distribution.
            First :math:`sample\_dim` elements contain the mean vector and the last
            :math:`sample\_dim ** 2` elements contain the covariance matrix row by row.
        n_samples : int
            Number of points to sample with each parameter in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples, sample_dim]
            Sample from multivariate normal distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        r"""Compute pdf associated to point.

        Parameters
        ----------
        point : array-like, shape=[..., sample_dim + sample_dim ** 2]
            Point representing a multivariate normal distribution.
            First :math:`sample\_dim` elements contain the mean vector and the last
            :math:`sample\_dim ** 2` elements contain the covariance matrix row by row.

        Returns
        -------
        pdf : function
            Probability density function of the multivariate normal
            distributions with parameters provided by point.
        """
        batch_shape = get_batch_shape(self.point_ndim, point)
        mean, cov = self._unstack_mean_covariance(point)
        det_cov = gs.linalg.det(cov)
        inv_cov = gs.linalg.inv(cov)
        pdf_normalization = 1 / gs.sqrt(gs.power(2 * gs.pi, self.sample_dim) * det_cov)

        def pdf(x):
            """Generate parameterized function for normal pdf.

            Parameters
            ----------
            x : array-like, shape=[n_samples, sample_dim]
                Points at which to compute the probability
                density function.
            """
            x_, (mean_, inv_cov_) = broadcast_to_multibatch(
                (x.shape[0],), batch_shape, x, mean, inv_cov
            )
            aux_0 = x_ - mean_
            aux_1 = gs.exp(-0.5 * gs.dot(aux_0, gs.matvec(inv_cov_, aux_0)))
            return gs.einsum(
                "...,...i->...i", pdf_normalization, gs.to_ndarray(aux_1, to_ndim=1)
            )

        return pdf


class UnivariateNormalToPoincareHalfSpaceDiffeo(Diffeo):
    """Diffeomorphism from univariate normal to Poincare half space."""

    def __call__(self, base_point):
        r"""Image of base point in the Poincare upper half-plane.

        This is the image by the diffeomorphism
        :math:`(mean, std) -> (mean, \sqrt{2} std)`.

        Parameters
        ----------
        base_point : array-like, shape=[..., 2]
            Point representing a normal distribution. Coordinates
            are mean and standard deviation.

        Returns
        -------
        image_point : array-like, shape=[..., 2]
            Image of base_point in the Poincare upper half-plane.
        """
        return gs.stack(
            [base_point[..., 0] / gs.sqrt(2.0), base_point[..., 1]], axis=-1
        )

    def inverse(self, image_point):
        r"""Inverse image of a point in the Poincare upper half-plane.

        This is the inverse image by the diffeomorphism
        :math:`(mean, std) -> (mean, \sqrt{2} std)`.

        Parameters
        ----------
        image_point : array-like, shape=[..., 2]
            Point in the upper half-plane.

        Returns
        -------
        base_point : array-like, shape=[..., 2]
            Inverse image of the image point, representing a normal
            distribution. Coordinates are mean and standard deviation.
        """
        return gs.stack(
            [image_point[..., 0] * gs.sqrt(2.0), image_point[..., 1]], axis=-1
        )

    def tangent(self, tangent_vec, base_point=None, image_point=None):
        r"""Image of tangent vector.

        This is the image by the tangent map of the diffeomorphism
        :math:`(mean, std) -> (mean, \sqrt{2} std)`.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 2]
            Tangent vector at base point.
        base_point : array-like, shape=[..., 2]
            Base point representing a normal distribution.
        image_point : array-like, shape=[..., *shape]
            Image point.

        Returns
        -------
        image_tangent_vec : array-like, shape=[..., 2]
            Image tangent vector at image of the base point.
        """
        return self(tangent_vec)

    def inverse_tangent(self, image_tangent_vec, image_point=None, base_point=None):
        r"""Inverse image of tangent vector.

        This is the inverse image by the tangent map of the diffeomorphism
        :math:`(mean, std) -> (mean, \sqrt{2} std)`.

        Parameters
        ----------
        image_tangent_vec : array-like, shape=[..., 2]
            Image of a tangent vector at image_point.
        image_point : array-like, shape=[..., 2]
            Image of a point representing a normal distribution.
        base_point : array-like, shape=[..., *shape]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., 2]
            Inverse image of image_tangent_vec.
        """
        return self.inverse(image_tangent_vec)


class UnivariateNormalMetric(PullbackDiffeoMetric):
    r"""Class for the Fisher information metric on univariate normal distributions.

    This is the pullback of the metric of the Poincare upper half-plane
    by the diffeomorphism :math:`(mean, std) -> (mean, \sqrt{2} std)`.
    """

    def __init__(self, space):
        diffeo = UnivariateNormalToPoincareHalfSpaceDiffeo()
        image_space = PoincareHalfSpace(dim=2)
        image_space.equip_with_metric(ScalarProductMetric(image_space, 2.0))
        super().__init__(space, diffeo, image_space)

    @staticmethod
    def metric_matrix(base_point):
        """Compute the metric matrix at the tangent space at base_point.

        Parameters
        ----------
        base_point : array-like, shape=[..., 2]
            Point representing a normal distribution (mean and scale).

        Returns
        -------
        mat : array-like, shape=[..., 2, 2]
            Metric matrix.
        """
        stds = base_point[..., 1]
        const = 1 / stds**2
        mat = gs.array([[1.0, 0], [0, 2]])
        return gs.einsum("...,ij->...ij", const, mat)

    def sectional_curvature(self, tangent_vec_a, tangent_vec_b, base_point=None):
        r"""Compute the sectional curvature.

        In the literature sectional curvature is noted K.

        For two orthonormal tangent vectors :math:`x,y` at a base point,
        the sectional curvature is defined by :math:`K(x,y) = <R(x, y)x, y>`.

        For non-orthonormal vectors, it is
        :math:`K(x,y) = <R(x, y)y, x> / (<x, x><y, y> - <x, y>^2)`.

        sectional_curvature(X, Y, P) = K(X,Y) where X, Y are tangent vectors
        at base point P.

        The information manifold of univariate normal distributions has constant
        sectional curvature given by :math:`K = - 1/2`.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., 2]
            Tangent vector at `base_point`.
        tangent_vec_b : array-like, shape=[..., 2]
            Tangent vector at `base_point`.
        base_point : array-like, shape=[..., 2]
            Point in the manifold.

        Returns
        -------
        sectional_curvature : array-like, shape=[...,]
            Sectional curvature at `base_point`.
        """
        sectional_curv = gs.array(-0.5)
        return repeat_out(
            self._space.point_ndim,
            sectional_curv,
            tangent_vec_a,
            tangent_vec_b,
            base_point,
        )


class CenteredNormalMetric:
    """Class for the Fisher information metric of centered normal distributions."""

    def __new__(cls, space):
        """Instantiate a scaled SPD affine metric."""
        space.equip_with_metric(SPDAffineMetric)
        return ScalarProductMetric(space, 1 / 2)


class DiagonalNormalMetric(RiemannianMetric):
    """Class for the Fisher information metric of diagonal normal distributions."""

    def __init__(self, space):
        super().__init__(space=space)
        self._univariate_normal = UnivariateNormalDistributions()

    def _stacked_mean_diagonal_to_1d_pairs(self, point, apply_sqrt=False):
        """Create pairs of 1d parameters from nd counterparts.

        Parameters
        ----------
        point: array-like, shape=[..., 2 * sample_dim]
            Stacked point (e.g. stacked means and diagonals).
        apply_sqrt: bool
            Determine if a square root is applied to the diagonals.

        Returns
        -------
        pairs : array-like, shape=[..., sample_dim, 2]
            Pairs of parameters (e.g. means and variances).
        """
        mean, diagonal = self._space._unstack_mean_diagonal(point)
        if apply_sqrt:
            diagonal = gs.sqrt(diagonal)
        return gs.stack([mean, diagonal], axis=-1)

    def _1d_pairs_to_stacked_mean_diagonal(self, point, apply_square=False):
        """Create nd stacked parameters from pairs of 1d counterparts.

        Parameters
        ----------
        pairs : array-like, shape=[..., sample_dim, 2]
            Pairs of parameters (e.g. means and variances).
        apply_square: bool
            Determine if a square is applied to the diagonals.

        Returns
        -------
        point: array-like, shape=[..., 2 * sample_dim]
            Stacked point (e.g. stacked means and diagonals).
        """
        mean = point[..., 0]
        diagonal = point[..., 1]
        if apply_square:
            diagonal = gs.power(diagonal, 2)
        return gs.concatenate([mean, diagonal], axis=-1)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., 2 * sample_dim]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., 2 * sample_dim]
            Tangent vector at base point.
        base_point: array-like, shape=[..., 2 * sample_dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        tangent_vec_a = self._stacked_mean_diagonal_to_1d_pairs(tangent_vec_a)
        tangent_vec_b = self._stacked_mean_diagonal_to_1d_pairs(tangent_vec_b)
        base_point = self._stacked_mean_diagonal_to_1d_pairs(
            base_point, apply_sqrt=True
        )
        inner_prod = self._univariate_normal.metric.inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        return gs.sum(inner_prod, axis=-1)

    def exp(self, tangent_vec, base_point):
        """Compute the Riemannian exponential.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 2 * sample_dim]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., 2 * sample_dim]
            Point.

        Returns
        -------
        end_point : array-like, shape=[..., 2 * sample_dim]
            Point reached by the geodesic starting from `base_point`
            with initial velocity `tangent_vec`.
        """
        tangent_vec = self._stacked_mean_diagonal_to_1d_pairs(tangent_vec)
        base_point = self._stacked_mean_diagonal_to_1d_pairs(
            base_point, apply_sqrt=True
        )
        end_point = self._univariate_normal.metric.exp(tangent_vec, base_point)
        return self._1d_pairs_to_stacked_mean_diagonal(end_point, apply_square=True)

    def log(self, point, base_point):
        """Compute Riemannian logarithm of a point wrt a base point.

        Parameters
        ----------
        point : array-like, shape=[..., 2 * sample_dim]
            Point.
        base_point : array-like, shape=[..., 2 * sample_dim]
            Point.

        Returns
        -------
        log : array-like, shape=[..., 2 * sample_dim]
            Tangent vector at the base point equal to the Riemannian logarithm
            of point at the base point.
        """
        point = self._stacked_mean_diagonal_to_1d_pairs(point, apply_sqrt=True)
        base_point = self._stacked_mean_diagonal_to_1d_pairs(
            base_point, apply_sqrt=True
        )
        log = self._univariate_normal.metric.log(point, base_point)
        return self._1d_pairs_to_stacked_mean_diagonal(log)

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        Parameters
        ----------
        base_point : array-like, shape=[..., 2 * sample_dim]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)


class UnivariateNormalDistributionsRandomVariable(ScipyUnivariateRandomVariable):
    """A univariate normal random variable."""

    def __init__(self, space):
        super().__init__(space, norm.rvs, norm.pdf)

    @staticmethod
    def _flatten_params(point, pre_flat_shape):
        loc = gs.expand_dims(point[..., 0], axis=-1)
        scale = gs.expand_dims(point[..., 1], axis=-1)

        flat_loc = gs.reshape(gs.broadcast_to(loc, pre_flat_shape), (-1,))
        flat_scale = gs.reshape(gs.broadcast_to(scale, pre_flat_shape), (-1,))
        return {"loc": flat_loc, "scale": flat_scale}


class SharedMeanNormalDistributionsRandomVariable(ScipyMultivariateRandomVariable):
    """A multivariate normal random variable with zero mean."""

    def __init__(self, space):
        rvs = lambda point, *args, **kwargs: multivariate_normal.rvs(
            *args, cov=point, **kwargs
        )
        pdf = lambda x, point: multivariate_normal.pdf(x, cov=point)
        super().__init__(space, rvs, pdf)


class DiagonalNormalDistributionsRandomVariable(ScipyMultivariateRandomVariable):
    """A diagonal multivariate normal random variable."""

    # TODO: update other children
    def __init__(self, space):
        rvs = lambda point, size: multivariate_normal.rvs(
            *space._unstack_mean_diagonal(point), size=size
        )
        pdf = lambda x, point: multivariate_normal.pdf(
            x, *space._unstack_mean_diagonal(point)
        )
        super().__init__(space, rvs, pdf)


class MultivariateNormalDistributionsRandomVariable(ScipyMultivariateRandomVariable):
    """A multivariate normal random variable."""

    def __init__(self, space):
        rvs = lambda point, size: multivariate_normal.rvs(
            *space._unstack_mean_covariance(point), size=size
        )
        pdf = lambda x, point: multivariate_normal.pdf(
            x, *space._unstack_mean_covariance(point)
        )
        super().__init__(space, rvs, pdf)



from geomstats.test_cases.geometry.base import ComplexMatrixVectorSpaceTestCase


class ComplexMatricesTestCase(ComplexMatrixVectorSpaceTestCase):
    def test_transconjugate(self, mat, expected, atol):
        res = self.space.transconjugate(mat)
        self.assertAllClose(res, expected, atol=atol)

    def test_is_hermitian(self, mat, expected, atol):
        res = self.space.is_hermitian(mat, atol=atol)
        self.assertAllEqual(res, expected)

    def test_is_hpd(self, mat, expected, atol):
        res = self.space.is_hpd(mat, atol=atol)
        self.assertAllEqual(res, expected)

    def test_to_hermitian(self, mat, expected, atol):
        res = self.space.to_hermitian(mat)
        self.assertAllClose(res, expected, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class ClusterInitializationTestCase(BaseEstimatorTestCase):
    @pytest.mark.random
    def test_initialization_belongs(self, n_samples, atol):
        X = self.data_generator.random_point(n_points=n_samples)
        init_cluster_centers = self.estimator._pick_init_cluster_centers(X)
        belongs = self.estimator.space.belongs(init_cluster_centers, atol=atol)
        self.assertTrue(gs.all(belongs))


class AgainstFrechetMeanTestCase(BaseEstimatorTestCase):
    @pytest.mark.random
    def test_against_frechet_mean(self, n_samples, atol):
        if self.estimator.n_clusters > 1:
            raise ValueError("Test only works for one cluster.")
        X = self.data_generator.random_point(n_points=n_samples)

        res = self.estimator.fit(X).cluster_centers_[0]
        res_ = self.other_estimator.fit(X).estimate_
        self.assertAllClose(res, res_, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.geometry.discrete_curves import SRVMetric
from geomstats.learning.frechet_mean import GradientDescent, variance
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test_cases.learning._base import (
    BaseEstimatorTestCase,
    MeanEstimatorMixinsTestCase,
)
from geomstats.vectorization import repeat_point


class FrechetMeanTestCase(MeanEstimatorMixinsTestCase, BaseEstimatorTestCase):
    @pytest.mark.random
    def test_logs_at_mean(self, atol):
        X = self.data_generator.random_point(n_points=2)

        mean = self.estimator.fit(X).estimate_

        logs = self.estimator.space.metric.log(X, mean)

        result = gs.linalg.norm(logs[1] + logs[0])
        self.assertAllClose(result, gs.array(0.0), atol)

    @pytest.mark.random
    def test_weighted_mean_two_points(self, atol):
        X = self.data_generator.random_point(n_points=2)
        weights = gs.random.rand(2)

        mean = self.estimator.fit(X, weights=weights).estimate_

        space = self.estimator.space
        expected = space.metric.exp(
            weights[1] / gs.sum(weights) * space.metric.log(X[1], X[0]), X[0]
        )
        self.assertAllClose(mean, expected, atol=atol)


class ElasticMeanTestCase(FrechetMeanTestCase):
    @pytest.mark.random
    def test_logs_at_mean(self, atol):
        space = self.estimator.space
        if not isinstance(space.metric, SRVMetric):
            return

        X = self.data_generator.random_point(n_points=2)

        mean = self.estimator.fit(X).estimate_

        logs = space.metric.log(X, mean)
        logs_srv = space.metric.diffeo.tangent(logs, base_point=mean)
        result = gs.linalg.norm(logs_srv[1] + logs_srv[0])

        self.assertAllClose(result, gs.array(0.0), atol)


class CircularMeanTestCase(FrechetMeanTestCase):
    @pytest.mark.random
    def test_against_optimization(self, n_points, atol):
        space = self.estimator.space
        X = self.data_generator.random_point(n_points)

        mean = self.estimator.fit(X).estimate_

        mean_gd = GradientDescent().minimize(space, points=X, weights=None)

        sum_sd_mean = gs.sum(space.metric.squared_dist(X, mean))
        sum_sd_mean_gd = gs.sum(space.metric.squared_dist(X, mean_gd))

        msg = f"circular mean: {mean}, {sum_sd_mean}\ngd: {mean_gd}, {sum_sd_mean_gd}"
        self.assertTrue(sum_sd_mean < sum_sd_mean_gd + atol, msg)


class VarianceTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def test_variance(self, points, base_point, expected, atol, weights=None):
        res = variance(self.space, points, base_point, weights=weights)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_variance_repeated_is_zero(self, n_samples, atol):
        base_point = point = self.data_generator.random_point(n_points=1)
        points = repeat_point(point, n_samples)

        self.test_variance(points, base_point, 0.0, atol)


class BatchGradientDescentTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    @pytest.mark.vec
    def test_against_default(self, n_points, n_reps, atol):
        points = self.data_generator.random_point(n_points)

        res_single = self.optimizer.minimize(self.space, points)

        rep_points = repeat_point(points, n_reps)
        rep_points = gs.moveaxis(rep_points, 0, 1)

        res = self.batch_optimizer.minimize(self.space, rep_points)

        self.assertAllClose(res, repeat_point(res_single, n_reps), atol)



import numpy as _np
import torch as _torch


def from_numpy(x):
    if _torch.is_tensor(x):
        return x
    return _torch.from_numpy(x)


def array(val, dtype=None):
    if _torch.is_tensor(val):
        if dtype is None or val.dtype == dtype:
            return val.clone()

        return cast(val, dtype=dtype)

    if isinstance(val, _np.ndarray):
        tensor = from_numpy(val)
        if dtype is not None and tensor.dtype != dtype:
            tensor = cast(tensor, dtype=dtype)

        return tensor

    if isinstance(val, (list, tuple)) and len(val):
        tensors = [array(tensor, dtype=dtype) for tensor in val]
        return _torch.stack(tensors)

    return _torch.tensor(val, dtype=dtype)


def cast(x, dtype):
    if _torch.is_tensor(x):
        return x.to(dtype=dtype)
    return array(x, dtype=dtype)



"""Class for the spider.

Lead authors: Anna Calissano & Jonas Lueg
"""

import geomstats.backend as gs
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.stratified.point_set import (
    Point,
    PointBatch,
    PointSet,
    PointSetMetric,
)
from geomstats.geometry.stratified.vectorization import broadcast_lists, vectorize_point


class SpiderPoint(Point):
    r"""Class for points of the Spider.

    A point in the Spider is :math:`(s,c) \in \mathbb{N} \times \mathbb{R}`.

    Parameters
    ----------
    stratum : int
        The stratum, an integer indicating the stratum the point lies in.
        If zero, then the point is on the origin.
    coord : array-like, shape=[1,]
        A positive number, the coordinate of the point. It must be zero if and
        only if the stratum is zero, i.e. the origin.
    """

    def __init__(self, stratum, coord):
        super().__init__()
        self.stratum = stratum
        self.coord = coord

    def __repr__(self):
        """Return a readable representation of the instance."""
        return f"r{self.stratum}: {self.coord[0]}"

    def _equal_single(self, point, atol=gs.atol):
        """Check equality against another point.

        Parameters
        ----------
        point : Point
            Point to compare against.
        atol : float

        Returns
        -------
        is_equal : bool
        """
        return self.stratum == point.stratum and abs(self.coord - point.coord) < gs.atol

    @vectorize_point((1, "point"))
    def equal(self, point, atol=gs.atol):
        """Check equality against another point.

        Parameters
        ----------
        point : Point or PointBatch
            Point to compare against.
        atol : float

        Returns
        -------
        is_equal : array-like, shape=[...]
        """
        return gs.array([self._equal_single(point_, atol) for point_ in point])


class Spider(PointSet):
    r"""Spider: a set of rays attached to the origin.

    The k-spider consists of k copies of the positive real line
    :math:`\mathbb{R}_{\geq 0}` glued together at the origin [Feragen2020]_.

    Parameters
    ----------
    n_rays : int
        Number of rays to attach to the origin.

    References
    ----------
    .. [Feragen2020]  Feragen, Aasa, and Tom Nye. "Statistics on stratified spaces."
        Riemannian Geometric Statistics in Medical Image Analysis.
        Academic Press, 2020. 299-342.
    """

    def __init__(self, n_rays, equip=True):
        super().__init__(equip=equip)
        self.n_rays = n_rays
        self.stratum_space = Euclidean(dim=1)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return SpiderMetric

    def random_point(self, n_samples=1):
        r"""Compute a random point of the spider set.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.

        Returns
        -------
        samples : SpiderPoint or PointBatch
            List of SpiderPoints randomly sampled from the Spider.
        """
        s = gs.random.randint(low=0, high=self.n_rays, size=(n_samples,))
        x = gs.abs(gs.random.normal(loc=10, scale=1, size=n_samples))
        random_point = [
            SpiderPoint(stratum=s[k], coord=gs.array([x[k]])) for k in range(n_samples)
        ]
        if n_samples == 1:
            return random_point[0]

        return PointBatch(random_point)

    @vectorize_point((1, "point"))
    def belongs(self, point, atol=gs.atol):
        r"""Check if a random point belongs to the spider set.

        Parameters
        ----------
        point : SpiderPoint or PointBatch
             Point to be checked.

        Returns
        -------
        belongs : array-like, shape=[...]
            Boolean evaluating if point belongs to the set.
        """
        results = []
        for point_ in point:
            results.append(
                self._coord_check(point_, atol) and self._n_rays_check(point_)
            )
        return gs.array(results)

    def _n_rays_check(self, point):
        r"""Check if a random point has the correct number of rays.

        Parameters
        ----------
        point : SpiderPoint
             Point to be checked.

        Returns
        -------
        belongs : bool
            Boolean denoting if the point has a ray in the rays set.
        """
        if point.stratum < self.n_rays:
            return True
        return False

    @staticmethod
    def _coord_check(point, atol=gs.atol):
        r"""Check if a random point has the correct length.

        Parameters
        ----------
        point : SpiderPoint
             Point to be checked.
        atol : float
            Absolute tolerance.

        Returns
        -------
        belongs : boolean
            Boolean denoting if the point has a positive length when on non-zero ray.
        """
        if point.coord <= -atol:
            return False
        return True


class SpiderMetric(PointSetMetric):
    """Geometry on the Spider, induced by the rays metric."""

    @property
    def _stratum_metric(self):
        return self._space.stratum_space.metric

    def _dist_single(self, point_a, point_b):
        """Compute the distance between two points on the Spider using the ray geometry.

        The spider metric is the metric in each ray extended to the Spider:
        given two points x, y on different rays, d(x, y) = d(x, 0) + d(0, y).

        Parameters
        ----------
        point_a : SpiderPoint
             Point in the Spider.
        point_b : SpiderPoint
             Point in the Spider.

        Returns
        -------
        dist : array-like, shape=[...]
            Distance between points.
        """
        if point_a.stratum == point_b.stratum:
            return self._stratum_metric.dist(point_a.coord, point_b.coord)

        return self._stratum_metric.dist(-point_a.coord, point_b.coord)

    @vectorize_point((1, "point_a"), (2, "point_b"))
    def dist(self, point_a, point_b):
        """Compute the distance between two points on the Spider using the ray geometry.

        The spider metric is the metric in each ray extended to the Spider:
        given two points x, y on different rays, d(x, y) = d(x, 0) + d(0, y).

        Parameters
        ----------
        point_a : SpiderPoint or PointBatch
             Point in the Spider.
        point_b : SpiderPoint or PointBatch
             Point in the Spider.

        Returns
        -------
        dist : array-like, shape=[...]
            Distance between points.
        """
        point_a, point_b = broadcast_lists(point_a, point_b)
        return gs.array(
            [
                self._dist_single(point_a_, point_b_)
                for point_a_, point_b_ in zip(point_a, point_b)
            ]
        )

    @vectorize_point((1, "initial_point"), (2, "end_point"))
    def geodesic(self, initial_point, end_point):
        """Return the geodesic between two lists of Spider points.

        Parameters
        ----------
        initial_point : SpiderPoint or PointBatch
             Point in the Spider.
        end_point : SpiderPoint or PointBatch
             Point in the Spider.

        Returns
        -------
        path : callable
            Return a vectorized geodesic function.
        """
        initial_point, end_point = broadcast_lists(initial_point, end_point)

        def _vec(t, fncs):
            if len(fncs) == 1:
                return fncs[0](t)

            return [fnc(t) for fnc in fncs]

        fncs = [
            self._geodesic_single(initial_point_, end_point_)
            for (initial_point_, end_point_) in zip(initial_point, end_point)
        ]
        return lambda t: _vec(t, fncs=fncs)

    def _geodesic_single(self, initial_point, end_point):
        """Compute the distance between two Spider points.

        Parameters
        ----------
        initial_point : SpiderPoint
             Point in the Spider.
        end_point : SpiderPoint
             Point in the Spider.

        Returns
        -------
        geo: function
            Geodesic between two Spider Points.
        """
        if initial_point.stratum == end_point.stratum:

            def ray_geo(t):
                ray_geod_func = self._stratum_metric.geodesic(
                    initial_point=initial_point.coord,
                    end_point=end_point.coord,
                )

                ray_geod_points = ray_geod_func(t)
                return PointBatch(
                    [
                        SpiderPoint(stratum=initial_point.stratum, coord=coord)
                        for coord in ray_geod_points
                    ]
                )

            return ray_geo

        def ray_geo(t):
            pseudo_ray_geod_func = self._stratum_metric.geodesic(
                initial_point=-initial_point.coord,
                end_point=end_point.coord,
            )
            pseudo_ray_geod_points = pseudo_ray_geod_func(t)
            return PointBatch(
                [
                    (
                        SpiderPoint(stratum=initial_point.stratum, coord=-coord)
                        if coord < 0.0
                        else SpiderPoint(stratum=end_point.stratum, coord=coord)
                    )
                    for coord in pseudo_ray_geod_points
                ]
            )

        return ray_geo



"""The n-dimensional hyperbolic space.

The n-dimensional hyperbolic space with Poincare ball model.

Lead author: Hadi Zaatiti.
"""

import math

import geomstats.algebra_utils as utils
import geomstats.backend as gs
from geomstats.geometry._hyperbolic import _Hyperbolic
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import repeat_out

EPSILON = 1e-6
NORMALIZATION_FACTOR_CST = gs.sqrt(gs.pi / 2)
PI_2_3 = gs.power(gs.array([2.0 * gs.pi]), gs.array([2 / 3]))
SQRT_2 = gs.sqrt(2.0)


class PoincareBall(_Hyperbolic, VectorSpaceOpenSet):
    """Class for the n-dimensional Poincare ball.

    Class for the n-dimensional Poincaré ball model. For other
    representations of hyperbolic spaces see the `Hyperbolic` class.

    Parameters
    ----------
    dim : int
        Dimension of the hyperbolic space.
    """

    def __init__(self, dim, equip=True):
        self.coords_type = "ball"
        super().__init__(
            dim=dim,
            embedding_space=Euclidean(dim),
            intrinsic=True,
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return PoincareBallMetric

    def belongs(self, point, atol=gs.atol):
        """Test if a point belongs to the hyperbolic space.

        Test if a point belongs to the hyperbolic space based on
        the poincare ball representation.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point to be tested.
        atol : float, optional
            Tolerance at which to evaluate how close the squared norm
            is to the reference value.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Array of booleans indicating whether the corresponding points
            belong to the hyperbolic space.
        """
        return gs.sum(point**2, axis=-1) < (1 - atol)

    def projection(self, point):
        """Project a point on the ball.

        Project a point by clipping such that l2
        norm being lower than 1

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in embedding Euclidean space.

        Returns
        -------
        projected_point : array-like, shape=[..., dim]
            Point projected on the ball.
        """
        l2_norm = gs.linalg.norm(point, axis=-1)
        if gs.any(l2_norm >= 1 - gs.atol):
            projected_point = gs.einsum(
                "...j,...->...j", point * (1 - gs.atol), 1.0 / l2_norm
            )
            return -gs.maximum(-projected_point, -point)

        return gs.copy(point)


class PoincareBallMetric(RiemannianMetric):
    """Class that defines operations using a Poincare ball."""

    def exp(self, tangent_vec, base_point):
        """Compute the Riemannian exponential of a tangent vector.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at a base point.
        base_point : array-like, shape=[..., dim]
            Point in the Poincare ball.

        Returns
        -------
        exp : array-like, shape=[..., dim]
            Point in the Poincare ball equal to the Riemannian exponential
            of tangent_vec at the base point.
        """
        squared_norm_bp = gs.sum(base_point**2, axis=-1)
        norm_tan = gs.linalg.norm(tangent_vec, axis=-1)
        lambda_base_point = 1 / (1 - squared_norm_bp)

        # This avoids dividing by 0
        norm_tan_eps = gs.where(gs.isclose(norm_tan, 0.0), EPSILON, norm_tan)
        direction = gs.einsum("...i,...->...i", tangent_vec, 1 / norm_tan_eps)

        factor = gs.tanh(lambda_base_point * norm_tan)

        return self.mobius_add(
            base_point, gs.einsum("...,...i->...i", factor, direction)
        )

    def log(self, point, base_point):
        """Compute Riemannian logarithm of a point wrt a base point.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point in the Poincare ball.
        base_point : array-like, shape=[..., dim]
            Point in the Poincare ball.

        Returns
        -------
        log : array-like, shape=[..., dim]
            Tangent vector at the base point equal to the Riemannian logarithm
            of point at the base point.
        """
        mobius_addition = self.mobius_add(-base_point, point)
        squared_norm_add = gs.sum(mobius_addition**2, axis=-1)
        squared_norm_bp = gs.sum(base_point**2, axis=-1)
        coef = (1 - squared_norm_bp) * utils.taylor_exp_even_func(
            squared_norm_add, utils.arctanh_card_close_0
        )
        return gs.einsum("...,...j->...j", coef, mobius_addition)

    def mobius_add(self, point_a, point_b, project_first=True):
        r"""Compute the Mobius addition of two points.

        The Mobius addition is useful to compute the log and exp in the
        'ball' representation.

        .. math::

            a\oplus b=\frac{(1+2\langle a,b\rangle + ||b||^2)a+
            (1-||a||^2)b}{1+2\langle a,b\rangle + ||a||^2||b||^2}

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            Point in Poincare ball.
        point_b : array-like, shape=[..., dim]
            Point in Poincare ball.
        project_first : boolean
            Project points on the ball or not (according to tolerance).

        Returns
        -------
        mobius_add : array-like, shape=[..., dim]
            Result of the Mobius addition.
        """
        if project_first:
            point_a = self._space.projection(point_a)
            point_b = self._space.projection(point_b)
        else:
            point_a_belong = self._space.belongs(point_a)
            point_b_belong = self._space.belongs(point_b)

            if not gs.all(point_a_belong) or not gs.all(point_b_belong):
                raise ValueError("Points do not belong to the Poincare ball")

        inner = gs.sum(point_a * point_b, axis=-1)
        squared_norm_a = gs.sum(point_a**2, axis=-1)
        squared_norm_b = gs.sum(point_b**2, axis=-1)
        num_1 = gs.einsum("...,...i->...i", 1 + 2 * inner + squared_norm_b, point_a)
        num_2 = gs.einsum("...,...i->...i", 1 - squared_norm_a, point_b)
        result = gs.einsum(
            "...,...i->...i",
            1.0 / (1 + 2 * inner + squared_norm_b * squared_norm_a),
            num_1 + num_2,
        )
        return self._space.projection(result)

    def squared_dist(self, point_a, point_b):
        """Squared geodesic distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            Point.
        point_b : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        sq_dist : array-like, shape=[...,]
            Squared distance.
        """
        return self.dist(point_a, point_b) ** 2

    def dist(self, point_a, point_b):
        """Compute the geodesic distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            First point in the Poincare ball.
        point_b : array-like, shape=[..., dim]
            Second point in the Poincare ball.

        Returns
        -------
        dist : array-like, shape=[...,]
            Geodesic distance between the two points.
        """
        point_a_norm = gs.clip(gs.sum(point_a**2, -1), 0.0, 1 - EPSILON)
        point_b_norm = gs.clip(gs.sum(point_b**2, -1), 0.0, 1 - EPSILON)

        diff_norm = gs.sum((point_a - point_b) ** 2, -1)
        norm_function = 1 + 2 * diff_norm / ((1 - point_a_norm) * (1 - point_b_norm))

        return gs.log(norm_function + gs.sqrt(norm_function**2 - 1))

    def retraction(self, tangent_vec, base_point):
        """Poincaré ball model retraction.

        Approximate the exponential map of the Poincare ball

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            vector in tangent space.
        base_point : array-like, shape=[..., dim]
            Second point in the Poincare ball.

        Returns
        -------
        point : array-like, shape=[..., dim]
            Retraction point.

        References
        ----------
        .. [1] Nickel et.al, "Poincaré Embedding for
            Learning Hierarchical Representation", 2017.
        """
        retraction_factor = (
            (1 - gs.sum(base_point**2, axis=-1, keepdims=True)) ** 2
        ) / 4

        return base_point - gs.einsum("...i,...j->...j", retraction_factor, tangent_vec)

    def metric_matrix(self, base_point=None):
        """Compute the inner product matrix.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, defaults to zeros if None.

        Returns
        -------
        inner_prod_mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.
        """
        if base_point is None:
            base_point = gs.zeros((1, self._space.dim))

        lambda_base = (2 / (1 - gs.sum(base_point * base_point, axis=-1))) ** 2
        identity = gs.eye(self._space.dim, self._space.dim)

        return gs.einsum("...,jk->...jk", lambda_base, identity)

    def normalization_factor(self, variances):
        """Return normalization factor of the Gaussian distribution.

        Parameters
        ----------
        variances : array-like, shape=[n,]
            Array of equally distant values of the
            variance precision time.

        Returns
        -------
        norm_func : array-like, shape=[n,]
            Normalisation factor for all given variances.
        """
        dim = self._space.dim
        binomial_coefficient = None
        n_samples = variances.shape[0]

        expand_variances = gs.expand_dims(variances, axis=0)
        expand_variances = gs.repeat(expand_variances, dim, axis=0)

        if binomial_coefficient is None:
            dim_range = gs.arange(dim)
            dim_range[0] = 1
            n_fact = dim_range.prod()

            k_fact = gs.concatenate(
                [
                    gs.expand_dims(dim_range[:i].prod(), 0)
                    for i in range(1, dim_range.shape[0] + 1)
                ],
                0,
            )

            nmk_fact = gs.flip(k_fact, 0)

            binomial_coefficient = n_fact / (k_fact * nmk_fact)

        binomial_coefficient = gs.expand_dims(binomial_coefficient, -1)
        binomial_coefficient = gs.repeat(binomial_coefficient, n_samples, axis=1)

        range_ = gs.expand_dims(gs.arange(dim), -1)
        range_ = gs.repeat(range_, n_samples, axis=1)

        ones_ = gs.expand_dims(gs.ones(dim), -1)
        ones_ = gs.repeat(ones_, n_samples, axis=1)

        alternate_neg = (-ones_) ** (range_)

        erf_arg = (((dim - 1) - 2 * range_) * expand_variances) / gs.sqrt(2)
        exp_arg = ((((dim - 1) - 2 * range_) * expand_variances) / gs.sqrt(2)) ** 2
        norm_func_1 = (1 + gs.erf(erf_arg)) * gs.exp(exp_arg)
        norm_func_2 = binomial_coefficient * norm_func_1
        norm_func_3 = alternate_neg * norm_func_2

        norm_func = (
            NORMALIZATION_FACTOR_CST
            * variances
            * norm_func_3.sum(0)
            * (1 / (2 ** (dim - 1)))
        )

        return norm_func

    def _compute_alpha(self, current_dim):
        """Compute factor used in normalization factor.

        Compute alpha factor given the two arguments.

        Parameters
        ----------
        current_dim : array-like, shape=[dim,]
            Array initialized at 0 with max range dim and step 1.
        """
        return (self._space.dim - 1 - 2 * current_dim) / SQRT_2

    def norm_factor_gradient(self, variances):
        """Compute normalization factor and its gradient.

        Compute normalization factor given current variance
        and dimensionality.

        Parameters
        ----------
        variances : array-like, shape=[n]
            Value of variance.

        Returns
        -------
        norm_factor : array-like, shape=[n]
            Normalisation factor.
        norm_factor_gradient : array-like, shape=[n]
            Gradient of the normalization factor.
        """
        dim = self._space.dim
        variances = gs.transpose(gs.to_ndarray(variances, to_ndim=2))
        dim_range = gs.arange(0, dim, 1.0)
        alpha = self._compute_alpha(dim_range)

        binomial_coefficient = gs.ones(dim)
        binomial_coefficient[1:] = (dim - 1 + 1 - dim_range[1:]) / dim_range[1:]
        binomial_coefficient = gs.cumprod(binomial_coefficient)

        beta = ((-gs.ones(dim)) ** dim_range) * binomial_coefficient

        sigma_repeated = gs.repeat(variances, dim, -1)
        prod_alpha_sigma = gs.einsum("ij,j->ij", sigma_repeated, alpha)
        term_2 = gs.exp((prod_alpha_sigma) ** 2) * (1 + gs.erf(prod_alpha_sigma))
        term_1 = gs.sqrt(gs.pi / 2.0) * (1.0 / (2 ** (dim - 1)))
        term_2 = gs.einsum("ij,j->ij", term_2, beta)
        norm_factor = term_1 * variances * gs.sum(term_2, axis=-1, keepdims=True)
        grad_term_1 = 1 / variances

        grad_term_21 = 1 / gs.sum(term_2, axis=-1, keepdims=True)

        grad_term_211 = (
            gs.exp((prod_alpha_sigma) ** 2)
            * (1 + gs.erf(prod_alpha_sigma))
            * gs.einsum("ij,j->ij", sigma_repeated, alpha**2)
            * 2
        )

        grad_term_212 = gs.repeat(
            gs.expand_dims((2 / gs.sqrt(gs.pi)) * alpha, axis=0),
            variances.shape[0],
            axis=0,
        )

        grad_term_22 = grad_term_211 + grad_term_212
        grad_term_22 = gs.einsum("ij, j->ij", grad_term_22, beta)
        grad_term_22 = gs.sum(grad_term_22, axis=-1, keepdims=True)

        norm_factor_gradient = grad_term_1 + (grad_term_21 * grad_term_22)

        return gs.squeeze(norm_factor), gs.squeeze(norm_factor_gradient)

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base
        point onto its image.
        In the case of the hyperbolic space, it does not depend on the base
        point and is infinite everywhere, because of the negative curvature.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)



"""Visualization for Geometric Statistics."""

import logging

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA

import geomstats.backend as gs
from geomstats.geometry.special_euclidean import SpecialEuclidean

SE2_GROUP = SpecialEuclidean(n=2, point_type="matrix")
SE2_VECT = SpecialEuclidean(n=2, point_type="vector")


class SpecialEuclidean2:
    """Class used to plot points in the 2d special euclidean group."""

    def __init__(self, points=None, point_type="matrix"):
        self.points = []
        self.point_type = point_type
        if points is not None:
            self.add_points(points)

    @staticmethod
    def set_ax(ax=None, x_lim=None, y_lim=None):
        if ax is None:
            ax = plt.subplot()
        if x_lim is not None:
            ax.set_xlim(x_lim)
        if y_lim is not None:
            ax.set_ylim(y_lim)
        return ax

    def add_points(self, points):
        if self.point_type == "vector":
            points = SE2_VECT.matrix_from_vector(points)
        if not gs.all(SE2_GROUP.belongs(points)):
            logging.warning("Some points do not belong to SE2.")
        if not isinstance(points, list):
            points = list(points)
        self.points.extend(points)

    def draw_points(self, ax, points=None, **kwargs):
        if points is None:
            points = gs.array(self.points)
        translation = points[..., :2, 2]
        frame_1 = points[:, :2, 0]
        frame_2 = points[:, :2, 1]
        ax.quiver(
            translation[:, 0],
            translation[:, 1],
            frame_1[:, 0],
            frame_1[:, 1],
            width=0.005,
            color="b",
        )
        ax.quiver(
            translation[:, 0],
            translation[:, 1],
            frame_2[:, 0],
            frame_2[:, 1],
            width=0.005,
            color="r",
        )
        ax.scatter(translation[:, 0], translation[:, 1], s=16, **kwargs)



from geomstats._backend._dtype_utils import _np_box_binary_scalar as _box_binary_scalar
from geomstats._backend._dtype_utils import _np_box_unary_scalar as _box_unary_scalar
from geomstats._backend._dtype_utils import (
    _pre_add_default_dtype_by_casting,
    _pre_allow_complex_dtype,
    _pre_cast_fout_to_input_dtype,
    _pre_cast_out_from_dtype,
    _pre_cast_out_to_input_dtype,
    _pre_set_default_dtype,
)

from .._backend_config import np_atol as atol
from .._backend_config import np_rtol as rtol
from ._dispatch import numpy as _np

_DTYPES = {
    _np.dtype("int32"): 0,
    _np.dtype("int64"): 1,
    _np.dtype("float32"): 2,
    _np.dtype("float64"): 3,
    _np.dtype("complex64"): 4,
    _np.dtype("complex128"): 5,
}

_COMPLEX_DTYPES = [
    _np.complex64,
    _np.complex128,
]


def is_floating(x):
    return x.dtype.kind == "f"


def is_complex(x):
    return x.dtype.kind == "c"


def is_bool(x):
    return x.dtype.kind == "b"


def as_dtype(value):
    """Transform string representing dtype in dtype."""
    return _np.dtype(value)


def _dtype_as_str(dtype):
    return dtype.name


def cast(x, dtype):
    return x.astype(dtype)


set_default_dtype = _pre_set_default_dtype(as_dtype)

_add_default_dtype_by_casting = _pre_add_default_dtype_by_casting(cast)
_cast_fout_to_input_dtype = _pre_cast_fout_to_input_dtype(cast, is_floating)
_cast_out_to_input_dtype = _pre_cast_out_to_input_dtype(
    cast, is_floating, is_complex, as_dtype, _dtype_as_str
)


_cast_out_from_dtype = _pre_cast_out_from_dtype(cast, is_floating, is_complex)
_allow_complex_dtype = _pre_allow_complex_dtype(cast, _COMPLEX_DTYPES)


def is_array(x):
    return type(x) is _np.ndarray


def to_ndarray(x, to_ndim, axis=0):
    x = _np.array(x)
    if x.ndim == to_ndim - 1:
        x = _np.expand_dims(x, axis=axis)

    if x.ndim != 0 and x.ndim < to_ndim:
        raise ValueError("The ndim was not adapted properly.")
    return x


def _get_wider_dtype(tensor_list):
    dtype_list = [_DTYPES.get(x.dtype, -1) for x in tensor_list]
    if len(dtype_list) == 1:
        return dtype_list[0], True

    wider_dtype_index = max(dtype_list)
    wider_dtype = list(_DTYPES.keys())[wider_dtype_index]

    return wider_dtype, False


def convert_to_wider_dtype(tensor_list):
    wider_dtype, same = _get_wider_dtype(tensor_list)
    if same:
        return tensor_list

    return [cast(x, dtype=wider_dtype) for x in tensor_list]


def _is_boolean(x):
    if isinstance(x, bool):
        return True
    if isinstance(x, (tuple, list)):
        return _is_boolean(x[0])
    if isinstance(x, _np.ndarray):
        return x.dtype == bool
    return False


def _is_iterable(x):
    if isinstance(x, (list, tuple)):
        return True
    if isinstance(x, _np.ndarray):
        return x.ndim > 0
    return False



from geomstats.test.test_case import TestCase


class LocalizationTestCase(TestCase):
    def test_propagate(self, state, sensor_input, expected, atol):
        res = self.model.propagate(state, sensor_input)
        self.assertAllClose(res, expected, atol=atol)

    def test_propagate_jacobian(self, state, sensor_input, expected, atol):
        res = self.model.propagation_jacobian(state, sensor_input)
        self.assertAllClose(res, expected, atol=atol)

    def test_observation_model(self, state, expected, atol):
        res = self.model.observation_model(state)
        self.assertAllClose(res, expected, atol=atol)

    def test_observation_jacobian(self, state, observation, expected, atol):
        res = self.model.observation_jacobian(state, observation)
        self.assertAllClose(res, expected, atol=atol)

    def test_innovation(self, state, observation, expected, atol):
        res = self.model.innovation(state, observation)
        self.assertAllClose(res, expected, atol=atol)


class NonLinearLocalizationTestCase(LocalizationTestCase):
    def test_preprocess_input(self, sensor_input, expected, atol):
        res = self.model.preprocess_input(sensor_input)
        for res_, expected_ in zip(res, expected):
            self.assertAllClose(res_, expected_, atol=atol)

    def test_rotation_matrix(self, theta, expected, atol):
        res = self.model.rotation_matrix(theta)
        self.assertAllClose(res, expected, atol=atol)

    def test_adjoint_map(self, state, expected, atol):
        res = self.model.adjoint_map(state)
        self.assertAllClose(res, expected, atol=atol)


class KalmanFilterTestCase(TestCase):
    def test_compute_gain(
        self, prior_values, process_values, obs_values, expected, atol
    ):
        self.estimator.initialize_covariances(prior_values, process_values, obs_values)

        res = self.estimator.compute_gain(None)
        self.assertAllClose(res, expected, atol=atol)

    def test_update(
        self,
        state,
        prior_values,
        process_values,
        obs_values,
        observation,
        expected_state,
        expected_cov,
        atol,
    ):
        self.estimator.state = state
        self.estimator.initialize_covariances(prior_values, process_values, obs_values)

        self.estimator.update(observation)
        self.assertAllClose(self.estimator.state, expected_state, atol=atol)
        self.assertAllClose(self.estimator.covariance, expected_cov, atol=atol)



"""Module for function spaces as geometric objects."""

import math

import geomstats.algebra_utils as utils
import geomstats.backend as gs
from geomstats.geometry.manifold import Manifold
from geomstats.geometry.riemannian_metric import RiemannianMetric


class HilbertSphere(Manifold):
    """Class for space of one dimensional functions with norm 1.

    The tangent space is given by functions that have
    zero inner-product with the base point

    Parameters
    ----------
    domain : array-like, shape=[n_samples,]
        Grid points on the domain.

    References
    ----------
    .. [Sea2016] Srivastava, Anuj, and Eric P. Klassen.
        Functional and shape data analysis.
        Vol. 1. New York: Springer, 2016.
    """

    def __init__(self, domain, equip=True):
        self.domain = domain
        super().__init__(
            dim=math.inf,
            shape=(len(domain),),
            equip=equip,
        )

        sinf_domain = (domain - min(domain)) / (max(domain) - min(domain))
        if gs.allclose(domain, sinf_domain):
            self._sinf = self
        else:
            self._sinf = HilbertSphere(sinf_domain)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return HilbertSphereMetric

    def projection(self, point):
        """Project a point to the infinite dimensional hypersphere.

        Parameters
        ----------
        point : array-like, shape=[..., n_samples]
            Discrete evaluation of a function.

        Returns
        -------
        projected_point : array-like, shape=[..., n_samples]
            Point projected to the hypersphere.
        """
        norm = self.metric.norm(point)
        return gs.einsum("...,...i->...i", 1.0 / norm, point)

    def belongs(self, point, atol=gs.atol):
        """Evaluate if the point belongs to the vector space.

        This method checks the shape of the input point.

        Parameters
        ----------
        point : array-like, shape=[..., n_samples]
            Point to test.
        atol : float

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space.
        """
        norms = self.metric.norm(point)

        return gs.isclose(norms, 1.0, atol)

    def is_tangent(self, vector, base_point, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., n_samples]
            Vector.
        base_point : array-like, shape=[..., n_samples]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        inner_product = self.metric.inner_product(vector, base_point)

        return gs.isclose(inner_product, gs.zeros_like(inner_product), atol)

    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., n_samples]
            Vector.
        base_point : array-like, shape=[..., n_samples]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n_samples]
            Tangent vector at base point.
        """
        sq_norm = self.metric.squared_norm(base_point)
        inner_product = self.metric.inner_product(vector, base_point)
        coef = inner_product / sq_norm
        return vector - gs.einsum("...,...j->...j", coef, base_point)

    def random_point(self, n_samples=1, bound=1.0):
        """Sample random points on the manifold.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            This parameter is ignored

        Returns
        -------
        samples : array-like, shape=[..., dim]
        """
        n_domain = len(self.domain)
        shape = (
            (n_domain,)
            if n_samples == 1
            else (
                n_samples,
                n_domain,
            )
        )
        point = gs.random.rand(*shape)

        return self.projection(point)


class HilbertSphereMetric(RiemannianMetric):
    """A Riemannian metric on the Hilbert sphere (functions of norm 1)."""

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute the inner-product of two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n_samples]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n_samples]
            Second tangent vector at base point.
        base_point : array-like, shape=[..., n_samples], optional
            Point on the hypersphere.

        Returns
        -------
        inner_prod : array-like, shape=[...,]
            Inner-product of the two tangent vectors.
        """
        tangent_vec_a, tangent_vec_b = gs.broadcast_arrays(tangent_vec_a, tangent_vec_b)
        x = gs.broadcast_to(self._space._sinf.domain, tangent_vec_a.shape)

        return gs.trapz(tangent_vec_a * tangent_vec_b, x=x, axis=-1)

    def exp(self, tangent_vec, base_point):
        """Compute the Riemannian exponential of a tangent vector.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n_samples]
            Tangent vector at a base point.
        base_point : array-like, shape=[..., n_samples]
            Point on the hypersphere.

        Returns
        -------
        exp : array-like, shape=[..., n_samples]
            Point on the hypersphere equal to the Riemannian exponential
            of tangent_vec at the base point.
        """
        proj_tangent_vec = self._space._sinf.to_tangent(tangent_vec, base_point)
        norm = self._space._sinf.metric.norm(proj_tangent_vec)

        norm = gs.clip(norm, -gs.pi, gs.pi)
        coef_1 = utils.taylor_exp_even_func(norm, utils.cos_close_0, order=4)
        coef_2 = utils.taylor_exp_even_func(norm, utils.sinc_close_0, order=4)
        exp = gs.einsum("...,...j->...j", coef_1, base_point) + gs.einsum(
            "...,...j->...j", coef_2, proj_tangent_vec
        )

        return exp

    def log(self, point, base_point):
        """Compute the Riemannian logarithm of a point.

        Parameters
        ----------
        point : array-like, shape=[..., n_samples]
            Point on the hypersphere.
        base_point : array-like, shape=[..., n_samples]
            Point on the hypersphere.

        Returns
        -------
        log : array-like, shape=[..., n_samples]
            Tangent vector at the base point equal to the Riemannian logarithm
            of point at the base point.
        """
        inner_prod = self.inner_product(base_point, point)
        cos_angle = gs.clip(inner_prod, -1.0, 1.0)
        theta = gs.arccos(cos_angle)
        coef_1_ = utils.taylor_exp_even_func(theta, utils.inv_sinc_close_0, order=5)
        coef_2_ = utils.taylor_exp_even_func(theta, utils.inv_tanc_close_0, order=5)
        log = gs.einsum("...,...j->...j", theta * coef_1_, point) - gs.einsum(
            "...,...j->...j", theta * coef_2_, base_point
        )

        return log



"""Torch based random backend."""

import torch as _torch
from torch import rand, randint
from torch.distributions.multivariate_normal import (
    MultivariateNormal as _MultivariateNormal,
)

from ._dtype import _allow_complex_dtype, _modify_func_default_dtype


def choice(x, a):
    """Generate a random sample from an array of given size."""
    if _torch.is_tensor(x):
        return x[_torch.randint(len(x), (a,))]

    return x


def seed(*args, **kwargs):
    return _torch.manual_seed(*args, **kwargs)


@_allow_complex_dtype
def normal(loc=0.0, scale=1.0, size=(1,)):
    if not hasattr(size, "__iter__"):
        size = (size,)
    return _torch.normal(mean=loc, std=scale, size=size)


def uniform(low=0.0, high=1.0, size=(1,), dtype=None):
    if not hasattr(size, "__iter__"):
        size = (size,)
    if low >= high:
        raise ValueError("Upper bound must be higher than lower bound")
    return (high - low) * rand(*size, dtype=dtype) + low


@_modify_func_default_dtype(copy=False, kw_only=True)
@_allow_complex_dtype
def multivariate_normal(mean, cov, size=(1,)):
    if not hasattr(size, "__iter__"):
        size = (size,)
    return _MultivariateNormal(mean, cov).sample(size)



"""Automatic differentiation in PyTorch."""

import functools

import torch as _torch
from torch.autograd.functional import hessian as _torch_hessian
from torch.autograd.functional import jacobian as _torch_jacobian


def _get_max_ndim_point(*points):
    """Identify point with higher dimension.

    Same as `geomstats.vectorization._get_max_ndim_point`.

    Parameters
    ----------
    points : array-like

    Returns
    -------
    max_ndim_point : array-like
        Point with higher dimension.
    """
    max_ndim_point = points[0]
    for point in points[1:]:
        if point.ndim > max_ndim_point.ndim:
            max_ndim_point = point

    return max_ndim_point


def _get_batch_shape(*points, point_ndims=1):
    """Get batch shape.

    Similar to `geomstats.vectorization.get_batch_shape`.

    Parameters
    ----------
    points : array-like or None
        Point belonging to the space.
    point_ndims : int or tuple[int]
        Point number of array dimensions.

    Returns
    -------
    batch_shape : tuple
        Returns the shape related with batch. () if only one point.
    """
    if isinstance(point_ndims, int):
        point_max_ndim = _get_max_ndim_point(*points)
        return point_max_ndim.shape[:-point_ndims]

    for point, point_ndim in zip(points, point_ndims):
        if point.ndim > point_ndim:
            return point.shape[:-point_ndim]

    return ()


def custom_gradient(*grad_funcs):
    """Create a decorator that allows a function to define its custom gradient(s).

    Parameters
    ----------
    *grad_funcs : callables
        Custom gradient functions.

    Returns
    -------
    decorator : callable
        This decorator, used on any function func, associates the
        input grad_funcs as the gradients of func.
    """

    def decorator(func):
        """Decorate a function to define its custome gradient(s).

        Parameters
        ----------
        func : callable
            Function whose gradients will be assigned by grad_funcs.

        Returns
        -------
        wrapped_function : callable
            Function func with gradients specified by grad_funcs.
        """

        class FuncWithGrad(_torch.autograd.Function):
            """Wrapper class for a function with custom grad."""

            @staticmethod
            def forward(ctx, *args, **kwargs):
                ctx.save_for_backward(*args)
                return func(*args, **kwargs)

            @staticmethod
            def backward(ctx, grad_output):
                inputs = ctx.saved_tensors

                if grad_output.ndim > 0:
                    return tuple(
                        (
                            _torch.einsum(
                                "n,n...->n...", grad_output, custom_grad(*inputs)
                            )
                            if input_.requires_grad
                            else None
                        )
                        for input_, custom_grad in zip(inputs, grad_funcs)
                    )

                return tuple(
                    grad_output * custom_grad(*inputs) if input_.requires_grad else None
                    for input_, custom_grad in zip(inputs, grad_funcs)
                )

        def wrapped_function(*args, **kwargs):
            return FuncWithGrad.apply(*args, **kwargs)

        return wrapped_function

    return decorator


def jacobian(func):
    """Return a function that returns the jacobian of func.

    Parameters
    ----------
    func : callable
        Function whose jacobian is computed.

    Returns
    -------
    _ : callable
        Function taking point as input and returning
        the jacobian of func at point.
    """

    def _jacobian(point):
        return _torch_jacobian(func=lambda x: func(x), inputs=point)

    return _jacobian


def jacobian_vec(func, point_ndim=1):
    """Return a function that returns the jacobian of func.

    We note that the jacobian function of torch is not vectorized
    by default, thus we modify its behavior here.

    Default pytorch behavior:

    If the jacobian for one point of shape (2,) is of shape (3, 2),
    then calling the jacobian on 4 points with shape (4, 2) will
    be of shape (3, 2, 4, 2).

    Modified behavior:

    Calling the jacobian on 4 points gives a tensor of shape (4, 3, 2).

    We use a for-loop to allow this function to be vectorized with
    respect to several inputs in point, because the flag vectorize=True
    fails.

    Parameters
    ----------
    func : callable
        Function whose jacobian is computed.

    Returns
    -------
    _ : callable
        Function taking point as input and returning
        the jacobian of func at point.
    """

    def _jacobian(point):
        if point.ndim == point_ndim:
            return _torch_jacobian(func=lambda x: func(x), inputs=point)
        return _torch.stack(
            [
                _torch_jacobian(func=lambda x: func(x), inputs=one_point)
                for one_point in point
            ],
            axis=0,
        )

    return _jacobian


def hessian(func, func_out_ndim=0):
    """Return a function that returns the hessian of func.

    Parameters
    ----------
    func : callable
        Function whose Hessian is computed.
    func_out_ndim : dim
        func output ndim.

    Returns
    -------
    _ : callable
        Function taking point as input and returning
        the hessian of func at point.
    """

    def _hessian(point):
        return _torch_hessian(func=lambda x: func(x), inputs=point, strict=True)

    def _hessian_vector_valued(point):
        def scalar_func(point, a):
            return func(point)[a]

        return _torch.stack(
            [
                hessian(functools.partial(scalar_func, a=a))(point)
                for a in range(func_out_ndim)
            ]
        )

    if func_out_ndim:
        return _hessian_vector_valued

    return _hessian


def hessian_vec(func, point_ndim=1, func_out_ndim=0):
    """Return a function that returns the hessian of func.

    We modify the default behavior of the hessian function of torch
    to return a tensor of shape (n_points, dim, dim) when several
    points are given as inputs.

    Parameters
    ----------
    func : callable
        Function whose Hessian is computed.
    func_out_ndim : dim
        func output ndim.

    Returns
    -------
    _ : callable
        Function taking point as input and returning
        the hessian of func at point.
    """
    hessian_func = hessian(func, func_out_ndim=func_out_ndim)

    def _hessian(point):
        if point.ndim == point_ndim:
            return hessian_func(point)
        return _torch.stack(
            [hessian_func(one_point) for one_point in point],
            axis=0,
        )

    return _hessian


def jacobian_and_hessian(func, func_out_ndim=0):
    """Return a function that returns func's jacobian and hessian.

    Parameters
    ----------
    func : callable
        Function whose jacobian and hessian
        will be computed. It must be real-valued.
    func_out_ndim : dim
        func output ndim.

    Returns
    -------
    func_with_jacobian_and_hessian : callable
        Function that returns func's jacobian and
        func's hessian at its inputs args.
    """

    def _jacobian_and_hessian(*args, **kwargs):
        """Return func's jacobian and func's hessian at args.

        Parameters
        ----------
        args : list
            Argument to function func and its gradients.
        kwargs : dict
            Keyword arguments to function func and its gradients.

        Returns
        -------
        jacobian : any
            Value of func's jacobian at input arguments args.
        hessian : any
            Value of func's hessian at input arguments args.
        """
        return jacobian(func)(*args), hessian(func, func_out_ndim=func_out_ndim)(*args)

    return _jacobian_and_hessian


def value_and_grad(func, argnums=0, point_ndims=1):
    """Return a function that returns func's value and gradients' values.

    Suitable for use in scipy.optimize.

    Parameters
    ----------
    func : callable
        Function whose value and gradient values
        will be computed. It must be real-valued.
    argnums: int or tuple[int]
        Specifies arguments to compute gradients with respect to.
    point_ndims: int or tuple[int]
        Specifies arguments ndim.

    Returns
    -------
    func_with_grad : callable
        Function that returns func's value and
        func's gradients' values at its inputs args.
    """
    argnums_ = (argnums,) if isinstance(argnums, int) else argnums

    def func_with_grad(*inputs, **kwargs):
        """Return func's value and func's gradients' values at args.

        Parameters
        ----------
        inputs : array-like, shape=[..., *point_shape]
        kwargs : dict
            Keyword arguments to function func and its gradients.

        Returns
        -------
        value : array-like, shape=[...,]
            Image of func at point.
        grad : array-like or tuple[array-like], shape=[..., *point_shape]
            Gradient of func at required points.
        """
        batch_shape = _get_batch_shape(*inputs, point_ndims=point_ndims)

        if len(inputs) > 1:
            point_ndims_ = (
                (point_ndims,) * len(inputs)
                if isinstance(point_ndims, int)
                else point_ndims
            )
            inputs_ = []
            for point, point_ndim in zip(inputs, point_ndims_):
                if point.shape[:-point_ndim] != batch_shape:
                    point = _torch.broadcast_to(point, batch_shape + point.shape)
                inputs_.append(point)
            inputs = inputs_

        inputs_ = []
        for index, point in enumerate(inputs):
            if index in argnums_ and not point.requires_grad:
                point = point.detach().requires_grad_(True)
            inputs_.append(point)
        inputs = inputs_
        value = func(*inputs, **kwargs).requires_grad_(True)

        if value.ndim > 0:
            sum_value = value.sum(axis=-1)
            sum_value.backward()
        else:
            value.backward()

        grads = tuple(
            point.grad.detach()
            for index, point in enumerate(inputs)
            if point.requires_grad and index in argnums_
        )
        if isinstance(argnums, int):
            grads = grads[0]
        return value.detach(), grads

    return func_with_grad


def value_and_jacobian(func):
    """Compute value and jacobian.

    NB: this is a naive implementation for consistency with autograd.

    Parameters
    ----------
    func : callable
        Function whose jacobian values will be computed.
    """

    def _value_and_jacobian(*args, **kwargs):
        """Return func's jacobian at args.

        Parameters
        ----------
        args : list
            Argument to function func and its gradients.
        kwargs : dict
            Keyword arguments to function func and its gradients.

        Returns
        -------
        value : array-like
            Value of func at input arguments args.
        jacobian : array-like
            Value of func's jacobian at input arguments args.
        """
        return (
            func(*args, **kwargs),
            jacobian_vec(func)(*args, **kwargs),
        )

    return _value_and_jacobian


def value_jacobian_and_hessian(func, func_out_ndim=0):
    """Compute value, jacobian and hessian.

    func is called as many times as the output dim.

    Parameters
    ----------
    func : callable
        Function whose jacobian and hessian values
        will be computed.
    func_out_ndim : int
        func output ndim.
    """

    def _value_jacobian_and_hessian(*args, **kwargs):
        """Return func's jacobian and func's hessian at args.

        Parameters
        ----------
        args : list
            Argument to function func and its gradients.
        kwargs : dict
            Keyword arguments to function func and its gradients.

        Returns
        -------
        value : array-like
            Value of func at input arguments args.
        jacobian : array-like
            Value of func's jacobian at input arguments args.
        hessian : array-like
            Value of func's hessian at input arguments args.
        """
        return (
            func(*args, **kwargs),
            jacobian_vec(func)(*args, **kwargs),
            hessian_vec(func, func_out_ndim=func_out_ndim)(*args, **kwargs),
        )

    return _value_jacobian_and_hessian



import geomstats.backend as gs
from geomstats.geometry.base import ImmersedSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.invariant_metric import BiInvariantMetric
from geomstats.geometry.pullback_metric import PullbackDiffeoMetric
from geomstats.test_cases.geometry.diffeo import CircleSO2Diffeo
from geomstats.test_cases.geometry.riemannian_metric import RiemannianMetricTestCase


class CircleAsSO2Metric(PullbackDiffeoMetric):
    def __init__(self, space, image_space):
        if not space.dim == 1:
            raise ValueError(
                "This dummy class using SO(2) metric for S1 has "
                "a meaning only when dim=1"
            )
        if not isinstance(image_space.metric, BiInvariantMetric):
            raise ValueError("Image space must be equipped with a bi-invariant metric")

        diffeo = CircleSO2Diffeo()
        super().__init__(space=space, diffeo=diffeo, image_space=image_space)


class CircleIntrinsic(ImmersedSet):
    def __init__(self, equip=True):
        super().__init__(dim=1, equip=equip)

    def immersion(self, point):
        return gs.hstack([gs.cos(point), gs.sin(point)])

    def _define_embedding_space(self):
        return Euclidean(dim=self.dim + 1)


class SphereIntrinsic(ImmersedSet):
    def __init__(self, equip=True):
        super().__init__(dim=2, equip=equip)

    def immersion(self, point):
        theta = point[..., 0]
        phi = point[..., 1]
        return gs.stack(
            [
                gs.cos(phi) * gs.sin(theta),
                gs.sin(phi) * gs.sin(theta),
                gs.cos(theta),
            ],
            axis=-1,
        )

    def _define_embedding_space(self):
        return Euclidean(dim=self.dim + 1)


class PullbackMetricTestCase(RiemannianMetricTestCase):
    def test_second_fundamental_form(self, base_point, expected, atol):
        res = self.space.metric.second_fundamental_form(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_mean_curvature_vector(self, base_point, expected, atol):
        res = self.space.metric.mean_curvature_vector(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_mean_curvature_vector_norm(self, base_point, expected, atol):
        mean_curvature = self.space.metric.mean_curvature_vector(base_point)
        res = gs.linalg.norm(mean_curvature)
        self.assertAllClose(res, expected, atol=atol)


class PullbackDiffeoMetricTestCase(RiemannianMetricTestCase):
    pass
